"""
Ce script sert à estimer la longueur des chemins d'arborescence des paramètres,
afin de ne pas avoir de chemins > 150 caractères (incompatible Windows).
Il est à utiliser avant de contribuer à l'harmonisation
"""


import os


def extract_paths_too_long(root_dir):
    root_dir = root_dir.rstrip(os.sep)

    with open("chemins_trop_longs.txt", "w") as outfile:
        for path, dirs, files in os.walk(root_dir):
            relative_path = path[len(root_dir) + 1:]
            for dir in dirs:
                if dir.startswith("."):
                    # Ignore hidden directories.
                    dirs.remove(dir)
            for file in files:
                if file.startswith("."):
                    # Ignore hidden files.
                    continue
                relative_file_path = os.path.join(relative_path, file)
                if len(relative_file_path) > 150:
                    outfile.write("{} here: {}\n".format(
                        len(relative_file_path),
                        relative_file_path,
                        ))


country_dir = os.path.dirname((os.path.dirname(os.path.abspath(__file__))))
parameters_dir = os.path.join(country_dir, "parameters")
extract_paths_too_long(parameters_dir)



import os

from ..model.base import *


dir_path = os.path.join(os.path.dirname(__file__), 'parameters')


# TODO: les baisses de charges n'ont pas été codées car annulées (toute ou en partie ?)
# par le Conseil constitutionnel

class plfr2014(Reform):
    name = 'Projet de Loi de Finances Rectificative 2014'

    class reduction_impot_exceptionnelle(Variable):
        definition_period = YEAR

        def formula_2013_01_01(foyer_fiscal, period, parameters):
            janvier = period.first_month

            nb_adult = foyer_fiscal('nb_adult', period)
            nb_parents = foyer_fiscal.declarant_principal.famille('nb_parents', period = janvier)
            rfr = foyer_fiscal('rfr', period)
            params = parameters(period).plfr2014.reduction_impot_exceptionnelle
            plafond = params.seuil * nb_adult + (nb_parents - nb_adult) * 2 * params.majoration_seuil
            montant = params.montant_plafond * nb_adult
            return min_(max_(plafond + montant - rfr, 0), montant)

    class reductions(Variable):
        label = "Somme des réductions d'impôt à intégrer pour l'année 2013"
        definition_period = YEAR

        def formula_2013_01_01(foyer_fiscal, period, parameters):
            accult = foyer_fiscal('accult', period)
            adhcga = foyer_fiscal('adhcga', period)
            cappme = foyer_fiscal('cappme', period)
            creaen = foyer_fiscal('creaen', period)
            daepad = foyer_fiscal('daepad', period)
            deffor = foyer_fiscal('deffor', period)
            dfppce = foyer_fiscal('dfppce', period)
            doment = foyer_fiscal('doment', period)
            domlog = foyer_fiscal('domlog', period)
            donapd = foyer_fiscal('donapd', period)
            duflot = foyer_fiscal('duflot', period)
            ecpess = foyer_fiscal('ecpess', period)
            garext = foyer_fiscal('garext', period)
            intagr = foyer_fiscal('intagr', period)
            invfor = foyer_fiscal('invfor', period)
            invlst = foyer_fiscal('invlst', period)
            ip_net = foyer_fiscal('ip_net', period)
            locmeu = foyer_fiscal('locmeu', period)
            mecena = foyer_fiscal('mecena', period)
            mohist = foyer_fiscal('mohist', period)
            patnat = foyer_fiscal('patnat', period)
            prcomp = foyer_fiscal('prcomp', period)
            reduction_impot_exceptionnelle = foyer_fiscal('reduction_impot_exceptionnelle', period)
            repsoc = foyer_fiscal('repsoc', period)
            resimm = foyer_fiscal('resimm', period)
            rsceha = foyer_fiscal('rsceha', period)
            saldom = foyer_fiscal('saldom', period)
            scelli = foyer_fiscal('scelli', period)
            sofica = foyer_fiscal('sofica', period)
            spfcpi = foyer_fiscal('spfcpi', period)
            total_reductions = accult + adhcga + cappme + creaen + daepad + deffor + dfppce + doment + domlog + \
                donapd + duflot + ecpess + garext + intagr + invfor + invlst + locmeu + mecena + mohist + patnat + \
                prcomp + repsoc + resimm + rsceha + saldom + scelli + sofica + spfcpi + reduction_impot_exceptionnelle
            return min_(ip_net, total_reductions)

    def apply(self):
        for variable in [self.reduction_impot_exceptionnelle, self.reductions]:
            self.update_variable(variable)
        self.modify_parameters(modifier_function = modify_parameters)


def modify_parameters(parameters):
    file_path = os.path.join(dir_path, 'plfr2014.yaml')
    plfr2014_parameters_subtree = load_parameter_file(name='plfr2014', file_path=file_path)

    file_path = os.path.join(dir_path, 'plfrss2014.yaml')
    plfrss2014_parameters_subtree = load_parameter_file(name='plfrss2014', file_path=file_path)

    parameters.add_child('plfr2014', plfr2014_parameters_subtree)
    parameters.add_child('plfrss2014', plfrss2014_parameters_subtree)
    return parameters



from numpy import fabs, timedelta64
from openfisca_france.model.base import Individu, Variable, MONTH, Enum, not_, ADD, \
    set_input_dispatch_by_period, set_input_divide_by_period, min_, date
from openfisca_france.model.caracteristiques_socio_demographiques.logement import TypesLieuResidence
from openfisca_france.model.revenus.activite.salarie import TypesContrat, TypesLieuEmploiFormation, \
    TypesCategoriesDemandeurEmploi


class aide_mobilite_date_demande(Variable):
    value_type = date
    default_value = date(1, 1, 1)
    entity = Individu
    label = "Date de demande d'évaluation à l'éligibilité de l'aide à la mobilité (AMOB) - (date du fait générateur)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"


class distance_activite_domicile(Variable):
    entity = Individu
    value_type = float
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"
    label = "Distance en kilomètres entre le lieu de l’entretien d’embauche, la reprise d’emploi, la formation, la prestation d’accompagnement, " \
            "l’immersion professionnelle (PMSMP), le concours public ou l’examen certifiant et le lieu de résidence du demandeur d'emploi"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class nombre_allers_retours(Variable):
    entity = Individu
    value_type = float
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"
    label = "Nombre d'aller/retour pour le calcul de l'aide à la mobilité de Pôle emploi - AMOB"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class aide_mobilite_duree_trajet(Variable):
    entity = Individu
    value_type = float
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"
    label = "Durée en minutes entre le lieu de l’entretien d’embauche, la reprise d’emploi, la formation, la prestation d’accompagnement, " \
            "l’immersion professionnelle (PMSMP), le concours public ou l’examen certifiant et le lieu de résidence du demandeur d'emploi"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class nuitees(Variable):
    entity = Individu
    value_type = int
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"
    label = "Nombre de nuitées pour le calcul de l'aide à la mobilité de Pôle emploi - AMOB"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class repas(Variable):
    entity = Individu
    value_type = int
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"
    label = "Nombre de repas pour le calcul de l'aide à la mobilité de Pôle emploi - AMOB"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class ContexteActivitePoleEmploi(Enum):
    __order__ = 'indetermine formation reprise_emploi recherche_emploi'  # Needed to preserve the enum order in Python 2
    indetermine = "Indeterminé"
    formation = "En formation"
    reprise_emploi = "En reprise d'emploi"
    recherche_emploi = "En recherche d'emploi"


class contexte_activite_pole_emploi(Variable):
    value_type = Enum
    possible_values = ContexteActivitePoleEmploi
    default_value = ContexteActivitePoleEmploi.indetermine
    entity = Individu
    label = "Le contexte d'activité au sens de Pôle Emploi"
    definition_period = MONTH


class DispositifsDeFormation(Enum):
    __order__ = 'autre bilan_competences permis_conduire_b accompagnement_creation_entreprise accompagnement_validation_acquis'  # Needed to preserve the enum order in Python 2
    autre = "Autre formation"
    bilan_competences = "Bilan de compétences"
    permis_conduire_b = "Permis de conduire B (code et/ou conduite)"
    accompagnement_creation_entreprise = "Accompagnement à la création d'entreprise"
    accompagnement_validation_acquis = "Accompagnement à la validation des acquis de l'experience (VAE)"


class dispositifs_formation(Variable):
    value_type = Enum
    possible_values = DispositifsDeFormation
    default_value = DispositifsDeFormation.autre
    entity = Individu
    label = "Dispositifs de formation"
    definition_period = MONTH


class formation_validee_pole_emploi(Variable):
    value_type = bool
    entity = Individu
    label = "La formation de l'individu est validée par Pôle emploi"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class formation_financee_ou_cofinancee(Variable):
    value_type = bool
    entity = Individu
    label = "La formation de l'individu est financée ou cofinancée (compte personnel de formation (CPF), fonds propres, Pôle Emploi, un tiers)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class TypesActiviteEnRechercheEmploi(Enum):
    __order__ = 'indeterminee entretien_embauche concours_public examen_certifiant prestation_accompagnement immersion_professionnelle_PMSMP'  # Needed to preserve the enum order in Python 2
    indeterminee = "Activité indéterminée"
    entretien_embauche = "Entretien d'embauche"
    concours_public = "Concours public"
    examen_certifiant = "Examen certifiant"
    prestation_accompagnement = "Prestation d'accompagnement"
    immersion_professionnelle_PMSMP = "Immersion professionnelle PMSMP"


class types_activite_en_recherche_emploi(Variable):
    value_type = Enum
    possible_values = TypesActiviteEnRechercheEmploi
    default_value = TypesActiviteEnRechercheEmploi.indeterminee
    entity = Individu
    label = "Les types d'activité dans un contexte de recherche d'emploi pour l'aide à la mobilité de Pôle Emploi - AMOB"
    definition_period = MONTH


class aide_mobilite_categories_demandeur_emploi_eligibles(Variable):
    value_type = bool
    entity = Individu
    label = "Le demandeur d'emploi appartient à une catégorie éligible pour l'aide à la mobilité - AMOB"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"

    def formula_2021_06_09(individu, period, parameters):

        pe_categorie_demandeur_emploi = individu('pole_emploi_categorie_demandeur_emploi', period)

        stagiaire_formation_professionnelle = individu('stagiaire', period)
        contrat_aide = individu('contrat_aide', period)

        categorie_4 = pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_4
        categorie_4_stagiaire_formation_professionnelle = categorie_4 * stagiaire_formation_professionnelle

        categorie_5 = pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_5
        categorie_5_contrat_aide = categorie_5 * contrat_aide

        categories_eligibles = ((pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_1)
                                + (pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_2)
                                + (pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_3)
                                + (categorie_4_stagiaire_formation_professionnelle + categorie_5_contrat_aide)
                                + (pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_6)
                                + (pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_7)
                                + (pe_categorie_demandeur_emploi == TypesCategoriesDemandeurEmploi.categorie_8))

        return categories_eligibles


class aide_mobilite_ressources_eligibles(Variable):
    value_type = bool
    entity = Individu
    label = "Le demandeur d'emploi perçoit une indemnité éligible pour l'aide à la mobilité - AMOB"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"

    def formula_2021_06_09(individu, period, parameters):
        epsilon = 0.0001
        lieu_de_residence = individu.menage('residence', period)
        mayotte = lieu_de_residence == TypesLieuResidence.mayotte
        hors_mayotte = not_(mayotte)

        allocation_individu = individu('allocation_retour_emploi_journaliere', period)
        allocations = parameters(period).chomage.allocation_retour_emploi
        allocation_minimale_hors_mayotte = allocations.montant_minimum_hors_mayotte * hors_mayotte
        allocation_minimale_mayotte = allocations.montant_minimum_mayotte * mayotte

        allocation_minimale_en_fonction_de_la_region = allocation_minimale_hors_mayotte + allocation_minimale_mayotte

        are_individu_egale_are_min = fabs(allocation_individu - allocation_minimale_en_fonction_de_la_region) < epsilon
        are_individu_inferieure_are_min = allocation_individu < allocation_minimale_en_fonction_de_la_region

        return are_individu_inferieure_are_min + are_individu_egale_are_min


class aide_mobilite_eligible(Variable):
    value_type = bool
    entity = Individu
    label = "Eligibilité à l'aide à la mobilité de Pôle Emploi - AMOB"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"
    documentation = '''
        1- L'individu doit être dans un contexte de recherche d'emploi, reprise d'emploi ou d'entrée en formation
            1.1 - Pour une formation ou une reprise d'emploi, la demande doit être faite au plus tard dans le mois suivant
            1.2 - Pour une recherche d'emploi, la demande doit être faite avant l'entretien d'embauche ou au plus tard dans un délai de 7 jours
        2- Sa reprise d'emploi ou son entretien d'emploi, cela doit concerner un cdi, cdd ou ctt de plus de 3 mois ou une formation de plus de 40h validée et financée
        3- L'individu est inscrit en catégorie 1, 2, 3, 4 "stagiaire de la formation professionnelle" ou 5 "contrat aidé", 6, 7 ou 8
        4- L'individu est non indemnisé ou son allocation est inférieure ou égale à l'ARE minimale
        5- L'emploi ou la formation se situe en France
        6-
            6.1 - Son "activité" doit être à plus de 60 km aller-retour de son lieu de résidence
            6.2 - Ou 20 km lorsque l'individu réside en dehors de la métropole
            6.3 - Ou 2 heures de trajet aller-retour
        7 - Les dispositifs suivants ne donne pas lieu au versement de l'aide à la mobilité :
            - le bilan de compétences
            - le permis de conduire B (code et/ou conduite)
            - l’accompagnement à la création d’entreprise
            - l’accompagnement à la validation des acquis de l’expérience (VAE)
    '''

    def formula_2021_06_09(individu, period, parameters):

        #  1
        contexte = individu('contexte_activite_pole_emploi', period)
        contrat_travail_debut = individu('contrat_de_travail_debut', period)
        date_debut_type_activite_recherche_emploi = individu('date_debut_recherche_emploi', period)
        contrat_de_travail_debut_en_mois = contrat_travail_debut.astype('M8[M]')
        amob_date_de_demande = individu("aide_mobilite_date_demande", period)
        parametres_amob = parameters(period).prestations_sociales.aide_mobilite
        date_contrat_limite_contexte_formation_reprise = min_((contrat_de_travail_debut_en_mois + 1) + (contrat_travail_debut - contrat_de_travail_debut_en_mois),
                                               (contrat_de_travail_debut_en_mois + 2) - timedelta64(1, 'D'))
        dates_demandes_amob_eligibles_formation_reprise = amob_date_de_demande <= date_contrat_limite_contexte_formation_reprise

        date_limite_contrat_contexte_recherche = date_debut_type_activite_recherche_emploi + (parametres_amob.delai_max - 1)  # 7 jours de date à date
        dates_demandes_amob_eligibles_recherche = amob_date_de_demande <= date_limite_contrat_contexte_recherche

        en_recherche_emploi = contexte == ContexteActivitePoleEmploi.recherche_emploi
        en_reprise_emploi = contexte == ContexteActivitePoleEmploi.reprise_emploi
        en_formation = contexte == ContexteActivitePoleEmploi.formation

        contextes_eligibles = (en_recherche_emploi * dates_demandes_amob_eligibles_recherche) \
            + ((en_reprise_emploi + en_formation) * dates_demandes_amob_eligibles_formation_reprise)

        #  2
        activite_en_recherche_emploi = individu('types_activite_en_recherche_emploi', period)
        reprises_emploi_types_activites = individu('contrat_de_travail_type', period)
        formation_validee = individu('formation_validee_pole_emploi', period)
        formation_financee = individu('formation_financee_ou_cofinancee', period)

        # Activites en recherche d'emploi
        en_entretien_embauche = (activite_en_recherche_emploi == TypesActiviteEnRechercheEmploi.entretien_embauche) * en_recherche_emploi

        activites_en_recherche_emploi_eligibles = not_((activite_en_recherche_emploi == TypesActiviteEnRechercheEmploi.entretien_embauche)
            + (activite_en_recherche_emploi == TypesActiviteEnRechercheEmploi.indeterminee)) \
            * en_recherche_emploi

        reprises_types_activites_formation = reprises_emploi_types_activites == TypesContrat.formation
        reprises_types_activites_cdi = reprises_emploi_types_activites == TypesContrat.cdi
        reprises_types_activites_cdd_ctt = (reprises_emploi_types_activites == TypesContrat.cdd) + (reprises_emploi_types_activites == TypesContrat.ctt)

        #  La formation doit être supérieure ou égale à 40 heures
        duree_formation = individu('duree_formation', period)
        periode_formation_eligible = duree_formation >= parametres_amob.duree_de_formation_minimum

        #  La durée de contrat de l'emploi doit être d'au moins 3 mois
        duree_de_contrat_3_mois_minimum = individu('contrat_de_travail_duree', period) >= parametres_amob.duree_de_contrat_minimum

        reprises_cdd_ctt_eligibles = reprises_types_activites_cdd_ctt * duree_de_contrat_3_mois_minimum

        types_et_duree_activite_eligibles = (((reprises_types_activites_cdi + reprises_cdd_ctt_eligibles) * (en_reprise_emploi + en_entretien_embauche))
                                            + (reprises_types_activites_formation
                                               * en_formation
                                               * formation_validee
                                               * formation_financee
                                               * periode_formation_eligible))

        activites_eligibles = (types_et_duree_activite_eligibles + activites_en_recherche_emploi_eligibles)

        #  3
        categories_eligibles = individu('aide_mobilite_categories_demandeur_emploi_eligibles', period)

        #  4
        montants_allocations_eligibles = individu('aide_mobilite_ressources_eligibles', period)

        #  5
        lieux_activite_eligibles = not_(individu('lieu_emploi_ou_formation', period) == TypesLieuEmploiFormation.non_renseigne)

        #  6
        lieu_de_residence = individu.menage('residence', period)
        temps_de_trajet = individu('aide_mobilite_duree_trajet', period)
        distance_aller_retour = individu('distance_activite_domicile', period) * 2

        temps_de_trajet_min = parametres_amob.duree_trajet_minimum
        reside_en_metropole = lieu_de_residence == TypesLieuResidence.metropole
        residence_renseignee = not_(lieu_de_residence == TypesLieuResidence.non_renseigne)

        distances_et_durees_aller_retour_eligibles = (((distance_aller_retour > parametres_amob.distance_minimum.metropole) * reside_en_metropole)
                                                    + ((distance_aller_retour > parametres_amob.distance_minimum.hors_metropole) * (not_(reside_en_metropole) * residence_renseignee))
                                                    + ((temps_de_trajet > temps_de_trajet_min) * residence_renseignee))

        #  7
        dispositifs_formations = individu('dispositifs_formation', period)
        dispositifs_formations_eligibles = dispositifs_formations == DispositifsDeFormation.autre

        eligibilite_amob = (contextes_eligibles
                       * activites_eligibles
                       * categories_eligibles
                       * montants_allocations_eligibles
                       * lieux_activite_eligibles
                       * distances_et_durees_aller_retour_eligibles
                       * dispositifs_formations_eligibles)

        return eligibilite_amob


class aide_mobilite(Variable):
    value_type = float
    entity = Individu
    label = "Montant de l'aide à la mobilité de Pôle emploi - AMOB"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"

    def formula_2021_06_09(individu, period, parameters):

        eligibilite_amob = individu('aide_mobilite_eligible', period)
        parametres_amob = parameters(period).prestations_sociales.aide_mobilite

        annee_glissante = period.start.period('year').offset(-1)

        aide_mobilite_12_derniers_mois = individu('aide_mobilite', annee_glissante, options=[ADD])

        montant_max = parametres_amob.montants.maximum
        montant_amob_deja_percu = min_(montant_max, fabs(aide_mobilite_12_derniers_mois))
        distance_aller_retour = individu('distance_activite_domicile', period)
        nb_aller_retour = individu('nombre_allers_retours', period)
        nb_nuitees = individu('nuitees', period)
        nb_repas = individu('repas', period)

        montant = parametres_amob.montants

        montants_frais_deplacement = montant.deplacement * distance_aller_retour * nb_aller_retour
        montants_frais_hebergement = montant.hebergement * nb_nuitees
        montants_frais_repas = montant.repas * nb_repas

        montants_theoriques = montants_frais_deplacement + montants_frais_hebergement + montants_frais_repas
        montants_max_attribuables = montant_max - montant_amob_deja_percu

        montants_reels = min_(montants_theoriques, montants_max_attribuables)

        return montants_reels * eligibilite_amob


class aide_mobilite_bon_de_transport(Variable):
    value_type = bool
    entity = Individu
    label = "Eligibilité à un bon de transport, part de l'aide à la mobilité de Pôle emploi (AMOB)"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "http://www.bo-pole-emploi.org/bulletinsofficiels/deliberation-n-2021-42-du-8-juin-2021-bope-n2021-43.html?type=dossiers/2021/bope-n-2021-043-du-11-juin-2021"

    def formula_2021_06_09(individu, period):

        contexte = individu('contexte_activite_pole_emploi', period)
        en_recherche_emploi = contexte == ContexteActivitePoleEmploi.recherche_emploi
        lieux_activite_eligibles = individu('lieu_emploi_ou_formation', period) == TypesLieuEmploiFormation.metropole_hors_corse
        categories_non_eligibles = not_(individu('aide_mobilite_categories_demandeur_emploi_eligibles', period))
        allocation_non_eligible = not_(individu('aide_mobilite_ressources_eligibles', period))

        return (en_recherche_emploi
            * lieux_activite_eligibles
            * categories_non_eligibles
            * allocation_non_eligible)



from numpy import datetime64, timedelta64

from openfisca_france.model.base import *
from openfisca_france.model.prelevements_obligatoires.prelevements_sociaux.cotisations_sociales.base import apply_bareme_for_relevant_type_sal


class jei_date_demande(Variable):
    value_type = date
    default_value = date(2099, 12, 31)
    entity = Individu
    label = "Date de demande (et d'octroi) du statut de jeune entreprise innovante (JEI)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class exoneration_cotisations_employeur_geographiques(Variable):
    value_type = float
    entity = Individu
    label = "Exonérations de cotisations employeur dépendant d'une zone géographique"
    reference = "https://www.apce.com/pid815/aides-au-recrutement.html?espace=1&tp=1"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        exoneration_cotisations_employeur_zfu = individu('exoneration_cotisations_employeur_zfu', period, options = [ADD])
        exoneration_cotisations_employeur_zrd = individu('exoneration_cotisations_employeur_zrd', period, options = [ADD])
        exoneration_cotisations_employeur_zrr = individu('exoneration_cotisations_employeur_zrr', period, options = [ADD])

        exonerations_geographiques = (
            exoneration_cotisations_employeur_zfu
            + exoneration_cotisations_employeur_zrd
            + exoneration_cotisations_employeur_zrr
            )

        return exonerations_geographiques


class exoneration_cotisations_employeur_jei(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour JEI (jeune entreprise innovante)"
    reference = "http://www.apce.com/pid1653/jeune-entreprise-innovante.html?pid=1653&pagination=2"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        jei_date_demande = individu('jei_date_demande', period)
        jeune_entreprise_innovante = individu('jeune_entreprise_innovante', period)
        plafond_securite_sociale = individu('plafond_securite_sociale', period)
        smic_proratise = individu('smic_proratise', period)
        categorie_salarie = individu('categorie_salarie', period)

        bareme_by_type_sal_name = parameters(period).cotsoc.cotisations_employeur
        bareme_names = ['vieillesse_deplafonnee', 'vieillesse_plafonnee', 'maladie', 'famille']

        exoneration = smic_proratise * 0.0
        for bareme_name in bareme_names:
            exoneration += apply_bareme_for_relevant_type_sal(
                bareme_by_type_sal_name = bareme_by_type_sal_name,
                bareme_name = bareme_name,
                categorie_salarie = categorie_salarie,
                base = min_(assiette_allegement, 4.5 * smic_proratise),
                plafond_securite_sociale = plafond_securite_sociale,
                round_base_decimals = 2,
                )

        exoneration_relative_year_passed = exoneration_relative_year(period, jei_date_demande)
        rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: 1,
            6: 1,
            7: 1,
            }  # TODO: move to parameters file
        for year_passed, rate in rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                exoneration[condition_on_year_passed] = rate * exoneration

        return - exoneration * jeune_entreprise_innovante


class exoneration_cotisations_employeur_zfu(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour l'embauche en ZFU (zone franche urbaine)"
    reference = "http://www.apce.com/pid553/exoneration-dans-les-zfu.html?espace=1&tp=1&pagination=2"
    definition_period = MONTH
    set_input = set_input_divide_by_period

# TODO
# Ce dispositif d'exonération sociale est fermé depuis le 1er janvier 2015 mais reste applicable aux entreprises qui
# en bénéficiaient avant cette date.
# - ne pas être détenues à plus de 25 % par des entreprises employant plus de 250 salariés et dont le chiffre d'affaires
#   ou dont le bilan excède 50 M€ ou 43 M€,
# - disposer des éléments d'exploitation ou des stocks nécessaires à l'activité des salariés,
# - être à jour de ses cotisations sociales ou avoir souscrit à un engagement d'apurement progressif de ses dettes.
#
# Secteurs d'activité concernés
#
# L'exonération est applicable, quel que soit le secteur d'activité.
# Toutefois, les entreprises exerçant une activité principale dans les secteurs de la construction automobile,
# construction navale, fabrication de fibres textiles artificielles ou synthétiques, sidérurgie ou des transports
# routiers de marchandises, ne pourront pas bénéficier de cette exonération.

# Embauche de résidents (clause d'embauche locale)
# Pour les entreprises qui se créent ou s'implantent dans une ZFU à compter du 1er janvier 2012,  le bénéfice de
# l'exonération des cotisations sociales est subordonnée lors de toute nouvelle embauche à la condition que la moitié
# de salariés embauchés ou employés résident en ZFU ou en zone urbaine sensible.
#
# Le respect de la condition d'embauche locale est apprécié à la date d'effet de la nouvelle embauche dès la deuxième
# embauche.
#
# Précision : les salariés employés sont ceux déjà présents dans l'entreprise à la date de la nouvelle embauche, les
# salariés embauchés sont ceux recrutés depuis la date de création ou d'implantation de l'entreprise en ZFU.
#
# Est considéré comme résident le salarié habitant soit dans la ZFU d'implantation, soit dans l'une des ZUS de l'unité
# urbaine où se trouve la ZFU. Le maire peut, à la demande de l'employeur, fournir des éléments d'informations relatifs
# à la qualité de résident dans la zone afin de déterminer si la proportion exigée est respectée.
#
# Si la proportion n'est pas respectée à la date d'effet de l'embauche, l'employeur dispose d'un délai de 3 mois pour
# régulariser la situation. A défaut, le bénéfice de l'exonération est suspendu du 1er jour du mois suivant
# l'expiration du délai de 3 mois, jusqu'au 1er jour du mois suivant la date où la condition est de nouveau remplie.
#
# Le salarié résident doit être titulaire d'un contrat à durée indéterminée ou d'un contrat à durée déterminée d'au
# moins 12 mois, conclu pour une durée minimale de 16 heures par semaine.
# 5 ans +
# Dans les entreprises de 5 salariés et plus, les cotisations employeur bénéficient d'un abattement sur la base
# imposable pendant 3 ans de :
# - 60 % la première année,
# - 40 % la seconde année,
# - 20 % la troisième année.
#
# Dans les entreprises de moins de 5 salariés, un abattement est appliqué sur 9 ans de la manière suivante :
# - 60 % les 5 premières années,
# - 40 % les 2 années suivantes,
# - 20 % les deux dernières années.
#
# Le cumul de l'ensemble des aides publiques de minimis (allégements fiscaux, sociaux et aides des collectivités
# territoriales) ne peut dépasser le plafond des aides de minimis, fixé à 200 000 euros sur une période glissante de 36
# mois (100 000 euros pour les entreprises de transport routier).

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        contrat_de_travail_type = individu('contrat_de_travail_type', period)
        TypesContrat = contrat_de_travail_type.possible_values
        contrat_de_travail_debut = individu('contrat_de_travail_debut', period)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', period)
        effectif_entreprise = individu('effectif_entreprise', period)
        entreprise_chiffre_affaire = individu('entreprise_chiffre_affaire', period)
        entreprise_bilan = individu('entreprise_bilan', period)
        taux_versement_transport = individu('taux_versement_transport', period)

        # TODO: move to parameters file
        entreprise_eligible = (entreprise_chiffre_affaire <= 1e7) | (entreprise_bilan <= 1e7)

        smic_proratise = individu('smic_proratise', period)
        zone_franche_urbaine = individu('zone_franche_urbaine', period)

        duree_cdd_eligible = (contrat_de_travail_fin > contrat_de_travail_debut + timedelta64(365, 'D'))
        # TODO: move to parameters file
        contrat_de_travail_eligible = (contrat_de_travail_debut <= datetime64("2014-12-31")) * ((contrat_de_travail_type == TypesContrat.cdi) + ((contrat_de_travail_type == TypesContrat.cdd) * (duree_cdd_eligible)))
        # TODO: move to parameters file

        eligible = (
            contrat_de_travail_eligible
            * (effectif_entreprise <= 50)
            * zone_franche_urbaine
            * entreprise_eligible
            )

        bareme_by_name = parameters(period).cotsoc.cotisations_employeur['prive_non_cadre']

        if period.start.year < 2007:
            fnal_contrib = parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.contribution_plus_de_10_salaries
            fnal_contrib_seuil = 10
        elif period.start.year >= 2007 and period.start.year < 2020:
            fnal_contrib = parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.contribution_plus_de_20_salaries
            fnal_contrib_seuil = 20
        else:
            fnal_contrib = parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.contribution_plus_de_50_salaries
            fnal_contrib_seuil = 50

        if period.start.year < 2019:
            taux_maladie = bareme_by_name['maladie'].rates[0]
        else:
            taux_maladie = 0

        taux_max = (
            bareme_by_name['vieillesse_deplafonnee'].rates[0]
            + bareme_by_name['vieillesse_plafonnee'].rates[0]
            + taux_maladie
            + bareme_by_name['famille'].rates[0]
            + parameters(period).prelevements_sociaux.autres_taxes_participations_assises_salaires.fnal.cotisation.rates[0]
            + fnal_contrib.rates[0] * (effectif_entreprise >= fnal_contrib_seuil)
            + taux_versement_transport
            )

        # TODO: move to parameters file : voir http://www.urssaf.fr/images/ref_lc2009-077.pdf
        seuil_max = 2
        seuil_min = 1.4

        taux_exoneration = compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min)
        exoneration_relative_year_passed = exoneration_relative_year(period, contrat_de_travail_debut)

        large_rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: .60,
            6: .40,
            7: .20,
            }  # TODO: move to parameters file

        small_rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: .60,
            6: .60,
            7: .60,
            8: .60,
            9: .60,
            10: .40,
            11: .40,
            12: .20,
            13: .20,
            }  # TODO: move to parameters file

        large_taux_exoneration = eligible * 0.0
        small_taux_exoneration = eligible * 0.0

        for year_passed, rate in large_rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                large_taux_exoneration[condition_on_year_passed] = rate * taux_exoneration

        for year_passed, rate in small_rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                small_taux_exoneration[condition_on_year_passed] = rate * taux_exoneration

        exoneration_cotisations_zfu = (
            eligible
            * assiette_allegement
            * (
                small_taux_exoneration
                * (effectif_entreprise <= 5)
                + large_taux_exoneration
                * (effectif_entreprise > 5)
                )
            )

        return exoneration_cotisations_zfu
        # TODO: propager dans le temps


class exoneration_cotisations_employeur_zrd(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour l'embauche en ZRD (zone de restructuration de la Défense)"
    reference = "http://www.apce.com/pid11668/exoneration-dans-les-zrd.html?espace=1&tp=1"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        entreprise_creation = individu('entreprise_creation', period)
        smic_proratise = individu('smic_proratise', period)
        zone_restructuration_defense = individu('zone_restructuration_defense', period)

        eligible = zone_restructuration_defense
        taux_max = .281  # TODO: move to parameters file
        seuil_max = 2.4
        seuil_min = 1.4
        taux_exoneration = compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min)

        exoneration_relative_year_passed = exoneration_relative_year(period, entreprise_creation)
        rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 2 / 3,
            4: 1 / 3,
            }  # TODO: move to parameters file
        ratio = eligible * 0.0
        for year_passed, rate in rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            if condition_on_year_passed.any():
                ratio[condition_on_year_passed] = rate

        exoneration_cotisations_zrd = ratio * taux_exoneration * assiette_allegement * eligible

        return exoneration_cotisations_zrd


class exoneration_cotisations_employeur_zrr(Variable):
    value_type = float
    entity = Individu
    label = "Exonération de cotisations employeur pour l'embauche en ZRR (zone de revitalisation rurale)"
    reference = "http://www.apce.com/pid538/embauches-en-zru-et-zrr.html?espace=1&tp=1"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    # Les entreprises et groupements d'employeurs exerçant une activité industrielle, commerciale, artisanale, agricole
    # ou libérale et cotisant au régime d'assurance chômage.
    # Les entreprises concernées, y compris chacune de celles appartenant à un groupement d'employeurs, doivent avoir
    # au moins un établissement situé en zone de revitalisation rurale.
    #
    # A noter : les associations à but non lucratif sont exclues du dispositif. Par contre, quelle que soit leur forme
    # juridique, les entreprises d'insertion ou d'intérim d'insertion peuvent en bénéficier. Les régies de quartier
    # peuvent en bénéficier lorsque leur activité est susceptible d'entraîner l'assujettissement à la TVA à l'impôt sur
    # les sociétés ainsi qu'à la contribution économique territoriale qu'elles en soient effectivement redevables
    # ou non.
    #
    # L'employeur ne doit avoir procédé à aucun licenciement économique durant les 12 mois précédant l'embauche.

    def formula(individu, period, parameters):
        assiette_allegement = individu('assiette_allegement', period)
        contrat_de_travail_type = individu('contrat_de_travail_type', period)
        TypesContrat = contrat_de_travail_type.possible_values
        contrat_de_travail_debut = individu('contrat_de_travail_debut', period)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', period)
        effectif_entreprise = individu('effectif_entreprise', period)
        smic_proratise = individu('smic_proratise', period)
        zone_revitalisation_rurale = individu('zone_revitalisation_rurale', period)

        duree_cdd_eligible = contrat_de_travail_fin > contrat_de_travail_debut + timedelta64(365, 'D')
        # TODO: move to parameters file
        contrat_de_travail_eligible = (contrat_de_travail_type == TypesContrat.cdi) + ((contrat_de_travail_type == TypesContrat.cdd) * (duree_cdd_eligible))

        duree_validite = (
            datetime64(period.start) + timedelta64(1, 'D') - contrat_de_travail_debut
            ).astype('timedelta64[Y]') < timedelta64(1, 'Y')

        eligible = (
            contrat_de_travail_eligible
            * (effectif_entreprise <= 50)
            * zone_revitalisation_rurale
            * duree_validite
            )

        taux_max = .281 if period.start.year < 2015 else .2655  # TODO: move to parameters file
        seuil_max = 2.4
        seuil_min = 1.5
        taux_exoneration = compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min)
        exoneration_cotisations_zrr = taux_exoneration * assiette_allegement * eligible

        return exoneration_cotisations_zrr


# Aides à la création
class exoneration_is_creation_zrr(Variable):
    value_type = float
    entity = Individu
    label = "Exonérations fiscales pour création d'une entreprise en zone de revitalisation rurale (ZRR)"
    reference = 'http://www.apce.com/pid11690/exonerations-d-impots-zrr.html?espace=1&tp=1'
    definition_period = YEAR
    calculate_output = calculate_output_divide

    def formula(individu, period, parameters):
        decembre = period.first_month.offset(11, 'month')
        effectif_entreprise = individu('effectif_entreprise', decembre)
        entreprise_benefice = individu('entreprise_benefice', period, options = [ADD])
        # TODO: MODIFIER avec création d'entreprise
        contrat_de_travail_type = individu('contrat_de_travail_type', decembre)

        TypesContrat = contrat_de_travail_type.possible_values

        contrat_de_travail_debut = individu('contrat_de_travail_debut', decembre)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', decembre)
        duree_eligible = contrat_de_travail_fin > contrat_de_travail_debut + timedelta64(365, 'D')
        # TODO: move to parameters file
        contrat_de_travail_eligible = (contrat_de_travail_type == TypesContrat.cdi) + ((contrat_de_travail_type == TypesContrat.cdd) * (duree_eligible))
        zone_revitalisation_rurale = individu('zone_revitalisation_rurale', decembre)

        eligible = (
            contrat_de_travail_eligible
            * (effectif_entreprise <= 50)
            * zone_revitalisation_rurale
            )

        exoneration_relative_year_passed = exoneration_relative_year(period, contrat_de_travail_debut)

        rate_by_year_passed = {
            0: 1,
            1: 1,
            2: 1,
            3: 1,
            4: 1,
            5: .75,
            6: .50,
            7: .25,
            }  # TODO: move to parameters file
        taux_exoneraion = eligible * 0.0

        for year_passed, rate in rate_by_year_passed.items():
            condition_on_year_passed = exoneration_relative_year_passed == timedelta64(year_passed, 'Y')
            taux_exoneraion[condition_on_year_passed] = rate

        return taux_exoneraion * entreprise_benefice
        # TODO: mettre sur toutes les années


# # class bassin_emploi_redynamiser(Variable):
#     value_type = bool
#     entity = Individu
#     label = "L'entreprise est située danns un bassin d'emploi à redynamiser(BER)"
#     # La liste des bassins d'emploi à redynamiser a été fixée par le décret n°2007-228 du 20 février 2007.
#     # Actuellement, deux régions sont concernées : Champagne-Ardenne (zone d'emploi de la Vallée de la Meuse)
#     # et Midi-Pyrénées (zone d'emploi de Lavelanet).
#
#     def formula(individu, period, parameters):
#         effectif_entreprise = individu('effectif_entreprise', period)
#         return (effectif_entreprise >= 1) * False

class jeune_entreprise_innovante(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est une jeune entreprise innovante"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        # Toute entreprise existante au 1er janvier 2004 ou créée entre le 1er janvier 2004 et le 31 décembre 2016 à
        # condition de remplir les conditions suivantes :
        #
        # avoir moins de 8 ans d'existence au moment de la demande
        #
        # être réellement nouvelle, c'est-à-dire ne pas avoir été créée dans le cadre d'une concentration,
        # d'une restructuration, d'une extension d'activité préexistante ou d'une reprise
        #
        # employer moins de 250 personnes au cours de l'exercice au titre duquel elle demande à bénéficier de ce statut
        #
        # réaliser un chiffre d'affaires inférieur à 50 M€  et disposer d'un total de bilan inférieur à 43 M€
        #
        # être indépendante, c'est-à-dire que son capital doit être détenu pour 50 % au minimum par :
        #
        # - des personnes physiques
        #
        # - une ou plusieurs autres JEI dont 50 % du capital au moins est détenu par des personnes physiques
        #
        # - des associations ou fondations reconnues d'utilité publique à caractère scientifique
        #
        # - des établissements de recherche et d'enseignement et leurs filiales
        #
        # - des structures d'investissement sous réserve qu'il n'y ait pas de lien de dépendance telles que des :
        #   -  fonds communs de placement dans l'innovation (FCPI)
        #   -  sociétés de capital-risque
        #   -  fonds d'investissement de proximité (FIP)
        #   -  sociétés de développement régional (SDR)
        #   -  sociétés financières d'innovation (SFI)
        #   -  sociétés unipersonnelles d'investissements à risques (SUIR).
        #
        # réaliser des dépenses de R§D représentant au moins 15 % des charges fiscalement déductibles au titre du même
        # exercice.
        effectif_entreprise = individu('effectif_entreprise', period)
        entreprise_bilan = individu('entreprise_bilan', period)
        entreprise_chiffre_affaire = individu('entreprise_chiffre_affaire', period)
        entreprise_creation = individu('entreprise_creation', period)
        # entreprise_depenses_rd =  individu('entreprise_depenses_rd', period)
        jei_date_demande = individu('jei_date_demande', period)
        # TODO: move to parameters file
        # entreprise_depenses_rd > .15 TODO
        independance = True

        jeune_entreprise_innovante = (
            independance
            * (effectif_entreprise < 250)
            * (entreprise_creation <= datetime64("2016-12-31"))
            * ((jei_date_demande + timedelta64(1, 'D') - entreprise_creation).astype('timedelta64[Y]') < timedelta64(8, 'Y'))
            * (entreprise_chiffre_affaire < 50e6)
            * (entreprise_bilan < 43e6)
            )

        return jeune_entreprise_innovante


class bassin_emploi_redynamiser(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située dans un bassin d'emploi à redynamiser (BER)"
    # La liste des bassins d'emploi à redynamiser a été fixée par le décret n°2007-228 du 20 février 2007.
    # Actuellement, deux régions sont concernées : Champagne-Ardenne (zone d'emploi de la Vallée de la Meuse)
    # et Midi-Pyrénées (zone d'emploi de Lavelanet).
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)

        return (effectif_entreprise >= 1) * False


class zone_restructuration_defense(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située dans une zone de restructuration de la Défense (ZRD)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)
        return (effectif_entreprise >= 1) * False


class zone_franche_urbaine(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située danns une zone franche urbaine (ZFU)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)
        return (effectif_entreprise >= 1) * False


class zone_revitalisation_rurale(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est située dans une zone de revitalisation rurale (ZRR)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        effectif_entreprise = individu('effectif_entreprise', period)
        return (effectif_entreprise >= 1) * False


# Helpers

def compute_taux_exoneration(assiette_allegement, smic_proratise, taux_max, seuil_max, seuil_min = 1):
    ratio_smic_salaire = smic_proratise / (assiette_allegement + 1e-16)
    # règle d'arrondi: 4 décimales au dix-millième le plus proche ( # TODO: reprise de l'allègement Fillon unchecked)
    return round_(
        taux_max * min_(1, max_(seuil_max * seuil_min * ratio_smic_salaire - seuil_min, 0) / (seuil_max - seuil_min)),
        4,
        )


def exoneration_relative_year(period, other_date):
    return (datetime64(period.start) + timedelta64(1, 'D') - other_date).astype('timedelta64[Y]')



"""Teste les cotisations par catégorie de salarié produites à l'issue de l'étape preprocessing."""

from collections import OrderedDict

from openfisca_core.parameters.parameter_scale import ParameterScale
from openfisca_france.model.prelevements_obligatoires.prelevements_sociaux.cotisations_sociales.base import (
    cotisations_employeur_by_categorie_salarie,
    cotisations_salarie_by_categorie_salarie,
    )


from .cache import tax_benefit_system


cotisations_salarie_by_name = {
    "agff": {
        "start_non_null_date": "2001-04-01",
        "final_null_date": "2019-01-01",
        },
    "agirc_arrco": {
        "start_non_null_date": "2019-01-01",
        },
    "agirc": {
        "start_non_null_date": "1948-01-01",
        "final_null_date": "2019-01-01",
        },
    "apec": {
        "start_non_null_date": "1971-01-01",
        },
    "arrco": {
        "start_non_null_date": "1962-01-01",
        "final_null_date": "2019-01-01",
        },
    "asf": {
        "start_non_null_date": "1984-01-01",
        },
    "ceg": {
        "start_non_null_date": "2019-01-01",
        },
    "cet": {
        "start_non_null_date": "1997-01-01",
        "final_null_date": "2019-01-01",
        },
    "cet2019": {
        "start_non_null_date": "2019-01-01",
        },
    "chomage": {
        "start_non_null_date": "1959-01-01",
        "final_null_date": "2018-10-01",
        },
    "cnracl_s_ti": {
        "start_non_null_date": "1947-09-19",
        },
    "cnracl_s_nbi": {
        "start_non_null_date": "1947-09-19",
        },
    "excep_solidarite": {
        "start_non_null_date": "1982-11-01",
        "final_null_date": "2018-01-01",
        },
    "ircantec": {
        "start_non_null_date": "1971-01-01",
        },
    "maladie_alsace_moselle": {
        "start_non_null_date": "1989-09-01",
        },
    "maladie": {
        "start_non_null_date": "1967-10-01",
        "final_null_date": "2018-01-01",
        },
    "pension": {
        "start_non_null_date": "1925-01-01",
        },
    "rafp": {
        "start_non_null_date": "2005-01-01"
        },
    "vieillesse_deplafonnee": {
        "start_non_null_date": "2004-07-01"
        },
    "vieillesse_plafonnee": {
        "start_non_null_date": "1993-07-01",
        },
    }


def test_preprocessing():
    """Tests the result of parameters preprocessing."""
    parameters = tax_benefit_system.parameters
    assert set(parameters.cotsoc.cotisations_employeur.children.keys()) == set([
        'prive_cadre',
        'prive_non_cadre',
        'public_non_titulaire',
        'public_titulaire_etat',
        'public_titulaire_hospitaliere',
        'public_titulaire_militaire',
        'public_titulaire_territoriale',
        ]), "Les barèmes de cotisations employeur de certaines catégories de salariés sont manquants"

    assert set(parameters.cotsoc.cotisations_salarie.children.keys()) == set([
        'prive_cadre',
        'prive_non_cadre',
        'public_non_titulaire',
        'public_titulaire_etat',
        'public_titulaire_hospitaliere',
        # 'public_titulaire_militaire',  FIXME Il y en a sûrement mais pas actuellement
        'public_titulaire_territoriale',
        ]), "Les barèmes de cotisations salarié de certaines catégories instant_sde salariés sont manquants"

    categorie_salaries = [
        "prive_cadre",
        "prive_non_cadre",
        "public_non_titulaire",
        "public_titulaire_etat",
        "public_titulaire_hospitaliere",
        "public_titulaire_territoriale",
        ]

    for categorie_salarie in categorie_salaries:
        test = parameters.cotsoc.cotisations_salarie.children[categorie_salarie].children.keys()
        target = cotisations_salarie_by_categorie_salarie[categorie_salarie]
        assert set(test) == set(target), "Les barèmes de cotisations salarié {} ne sont pas les bons".format(
            categorie_salarie)

        test = parameters.cotsoc.cotisations_employeur.children[categorie_salarie].children.keys()
        target = cotisations_employeur_by_categorie_salarie[categorie_salarie]
        assert set(test) == set(target), "Les barèmes de cotisations employeur {} ne sont pas les bons".format(
            categorie_salarie)

    for categorie_salarie in categorie_salaries:
        cotisations_salaries = set(cotisations_salarie_by_categorie_salarie[categorie_salarie])
        for cotisation_salarie in sorted(cotisations_salaries):
            bareme = parameters.cotsoc.cotisations_salarie.children[categorie_salarie].children[cotisation_salarie]

            assert isinstance(bareme, ParameterScale), f"{cotisation_salarie} entry's class is {type(bareme)} which is not a ParameterScale"

            final_null_date = cotisations_salarie_by_name[cotisation_salarie].get("final_null_date")
            if final_null_date:
                thresholds = [
                    dict(
                        (parameter_at_instant.instant_str, parameter_at_instant.value)
                        for parameter_at_instant in bracket.threshold.values_list
                        )
                    for bracket in bareme.brackets
                    ]
                final_thresholds_by_instant_str = OrderedDict(sorted(threshold.items(), reverse = True)[0] for threshold in thresholds)
                assert all([final_threshold is None for final_threshold in final_thresholds_by_instant_str.values()]), "Barème salarié {} ne s'éteint pas (il devrait en {})".format(
                    cotisation_salarie,
                    final_null_date,
                    )
                assert max(final_thresholds_by_instant_str.keys()) == final_null_date, "Barème salarié {} ne s'éteint pas en {}".format(
                    cotisation_salarie,
                    final_null_date,
                    )

            start_non_null_date = cotisations_salarie_by_name[cotisation_salarie].get("start_non_null_date")
            if start_non_null_date:
                thresholds = [
                    dict(
                        (parameter_at_instant.instant_str, parameter_at_instant.value)
                        for parameter_at_instant in bracket.threshold.values_list
                        )
                    for bracket in bareme.brackets
                    ]
                start_thresholds_by_instant_str = OrderedDict(sorted(threshold.items())[0] for threshold in thresholds)
                assert all([start_threshold is not None for start_threshold in start_thresholds_by_instant_str.values()]), "Barème salarié {} ne commence pas à la bonne date (il devrait en {})".format(
                    cotisation_salarie,
                    start_non_null_date,
                    )
                assert min(start_thresholds_by_instant_str.keys()) == start_non_null_date, "Barème salarié {} ne commence pas en {}".format(
                    cotisation_salarie,
                    start_non_null_date,
                    )



from numpy.core.defchararray import startswith

from openfisca_france.model.base import *


class coloc(Variable):
    value_type = bool
    entity = Menage
    label = "Vie en colocation"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class logement_crous(Variable):
    value_type = bool
    entity = Menage
    label = "Le logement est géré par les CROUS "
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class logement_chambre(Variable):
    value_type = bool
    entity = Menage
    label = "Le logement est considéré comme une chambre"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class loyer(Variable):
    value_type = float
    entity = Menage
    set_input = set_input_divide_by_period
    label = "Loyer ou mensualité d'emprunt pour un primo-accédant"
    definition_period = MONTH


class depcom(Variable):
    value_type = str
    max_length = 5
    entity = Menage
    label = "Code INSEE (depcom) du lieu de résidence"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class charges_locatives(Variable):
    value_type = float
    entity = Menage
    set_input = set_input_divide_by_period
    label = 'Charges locatives'
    definition_period = MONTH


class proprietaire_proche_famille(Variable):
    value_type = bool
    entity = Famille
    label = "Le propriétaire du logement a un lien de parenté avec la personne de référence ou son conjoint"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class habite_chez_parents(Variable):
    value_type = bool
    entity = Individu
    label = "L'individu habite chez ses parents"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class statut_occupation_logement(Variable):
    value_type = Enum
    possible_values = TypesStatutOccupationLogement  # defined in model/base.py
    entity = Menage
    default_value = TypesStatutOccupationLogement.non_renseigne
    label = "Statut d'occupation du logement"
    set_input = set_input_dispatch_by_period
    definition_period = MONTH


class residence_ile_de_france(Variable):
    label = "Le logement est situé dans la région Île-de-France"
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return sum([startswith(depcom, str.encode(departement_idf)) for departement_idf in parameters(period).geopolitique.departements_idf])  # TOOPTIMIZE: string encoding into bytes array should be done at load time


class residence_dom(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period):
        residence_guadeloupe = menage('residence_guadeloupe', period)
        residence_martinique = menage('residence_martinique', period)
        residence_guyane = menage('residence_guyane', period)
        residence_reunion = menage('residence_reunion', period)
        residence_mayotte = menage('residence_mayotte', period)

        return residence_guadeloupe + residence_martinique + residence_reunion + residence_guyane + residence_mayotte


class residence_guadeloupe(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'971')


class residence_martinique(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'972')


class residence_guyane(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'973')


class residence_reunion(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'974')


class residence_saint_pierre_et_miquelon(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'975')


class residence_mayotte(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'976')


class residence_saint_bartelemy(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'977')


class residence_saint_martin(Variable):
    value_type = bool
    entity = Menage
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        depcom = menage('depcom', period)
        return startswith(depcom, b'978')


class TypesLieuResidence(Enum):
    non_renseigne = "Non renseigné"
    metropole = "Métropole"
    guadeloupe = "Guadeloupe"
    martinique = "Martinique"
    guyane = "Guyane"
    la_reunion = "La réunion"
    saint_pierre_et_miquelon = "Saint Pierre et Miquelon"
    mayotte = "Mayotte"
    saint_bartelemy = "Saint Bartelemy"
    saint_martin = "Saint Martin"


class residence(Variable):
    value_type = Enum
    possible_values = TypesLieuResidence
    default_value = TypesLieuResidence.non_renseigne
    entity = Menage
    label = "Zone de résidence"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(menage, period, parameters):
        return select(
            [
                menage('residence_guadeloupe', period),
                menage('residence_martinique', period),
                menage('residence_guyane', period),
                menage('residence_reunion', period),
                menage('residence_saint_pierre_et_miquelon', period),
                menage('residence_mayotte', period),
                menage('residence_saint_bartelemy', period),
                menage('residence_saint_martin', period)
                ],
            [
                TypesLieuResidence.guadeloupe,
                TypesLieuResidence.martinique,
                TypesLieuResidence.guyane,
                TypesLieuResidence.la_reunion,
                TypesLieuResidence.saint_pierre_et_miquelon,
                TypesLieuResidence.mayotte,
                TypesLieuResidence.saint_bartelemy,
                TypesLieuResidence.saint_martin
                ],
            default=TypesLieuResidence.metropole
            )



from typing import Callable, Tuple

import numpy
import pytest

from ..cache import tax_benefit_system
from openfisca_core.reforms import Reform
from openfisca_core.simulations import Simulation
from openfisca_france.reforms.plf2016 import (
    plf2016,
    plf2016_counterfactual,
    plf2016_counterfactual_2014,
    )


@pytest.fixture
def simulations(new_scenario) -> Callable[..., Tuple[Simulation]]:
    def _simulations(reform_class: Reform, year: int) -> Tuple[Simulation]:
        scenario = new_scenario(
            reform = reform_class(tax_benefit_system),
            count = 2,
            _max = 18000,
            _min = 0,
            name = "salaire_imposable",
            year = year,
            people = 1,
            )

        return scenario.new_simulation(use_baseline = True), scenario.new_simulation()

    return _simulations


def equal(before: numpy.ndarray, after: numpy.ndarray) -> bool:
    return numpy.all(numpy.equal(before, after))


def not_equal(before: numpy.ndarray, after: numpy.ndarray) -> bool:
    return numpy.any(numpy.not_equal(before, after))


@pytest.mark.parametrize(
    "reform_class, expected", (
        (plf2016, equal),
        (plf2016_counterfactual, not_equal),
        (plf2016_counterfactual_2014, not_equal),
        ),
    )
def test_plf2016(reform_class, expected, simulations, year = 2015):
    actual_simulation, reform_simulation = simulations(reform_class, year)
    actual_impots = actual_simulation.calculate('impots_directs', year)
    reform_impots = reform_simulation.calculate('impots_directs', year)
    assert expected(actual_impots, reform_impots)



from openfisca_france.model.base import *
from openfisca_france.model.prestations.prestations_familiales.base_ressource import nb_enf


class ars(Variable):
    value_type = float
    entity = Famille
    label = "Allocation de rentrée scolaire"
    reference = "http://vosdroits.service-public.fr/particuliers/F1878.xhtml"
    definition_period = YEAR

    def formula(famille, period, parameters):
        '''
        Allocation de rentrée scolaire brute de CRDS
        '''
        janvier = period.first_month
        septembre = period.start.offset('first-of', 'year').offset(9, 'month').period('month')
        af_nbenf = famille('af_nbenf', septembre)
        base_ressources = famille('prestations_familiales_base_ressources', janvier)
        ars = parameters(septembre).prestations_sociales.prestations_familiales.education_presence_parentale.ars
        # TODO: convention sur la mensualisation
        # On tient compte du fait qu'en cas de léger dépassement du plafond, une allocation dégressive
        # (appelée allocation différentielle), calculée en fonction des revenus, peut être versée.

        bmaf = parameters(septembre).prestations_sociales.prestations_familiales.bmaf.bmaf
        # On doit prendre l'âge en septembre
        enf_05 = nb_enf(famille, septembre, ars.ars_cond.age_entree_primaire - 1, ars.ars_cond.age_entree_primaire - 1)  # 5 ans et 6 ans avant le 31 décembre
        # enf_05 = 0
        # Un enfant scolarisé qui n'a pas encore atteint l'âge de 6 ans
        # avant le 1er février 2012 peut donner droit à l'ARS à condition qu'il
        # soit inscrit à l'école primaire. Il faudra alors présenter un
        # certificat de scolarité.
        enf_primaire = enf_05 + nb_enf(famille, septembre, ars.ars_cond.age_entree_primaire, ars.ars_cond.age_entree_college - 1)
        enf_college = nb_enf(famille, septembre, ars.ars_cond.age_entree_college, ars.ars_cond.age_entree_lycee - 1)
        enf_lycee = nb_enf(famille, septembre, ars.ars_cond.age_entree_lycee, ars.ars_cond.age_sortie_lycee)

        arsnbenf = enf_primaire + enf_college + enf_lycee

        # Plafond en fonction du nb d'enfants A CHARGE (Cf. article R543)
        ars_plaf_res = ars.ars_plaf.plafond_ressources * (1 + af_nbenf * ars.ars_plaf.majoration_par_enf_supp)

        arsbase = bmaf * (
            ars.ars_m.taux_primaire * enf_primaire
            + ars.ars_m.taux_college * enf_college
            + ars.ars_m.taux_lycee * enf_lycee
            )

        # Forme de l'ARS  en fonction des enfants a*n - (rev-plaf)/n
        # ars_diff = (ars_plaf_res + arsbase - base_ressources) / arsnbenf
        ars_montant = (arsnbenf > 0) * max_(0, arsbase - max_(0, (base_ressources - ars_plaf_res) / max_(1, arsnbenf)))

        return ars_montant * (ars_montant >= ars.montant_minimum_verse)



from openfisca_france.model.base import *


class chomeur_longue_duree(Variable):
    cerfa_field = {
        0: "1AI",
        1: "1BI",
        2: "1CI",
        3: "1DI",
        4: "1EI",
        }
    value_type = bool
    entity = Individu
    label = "Demandeur d'emploi inscrit depuis plus d'un an"
    definition_period = YEAR
    # Pour toutes les variables de ce type, les pac3 ne sont plus proposés après 2007


class chomage_brut(Variable):
    value_type = float
    entity = Individu
    label = "Chômage brut (revenus de remplacement pour les demandeurs d'emploi)"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    calculate_output = calculate_output_add

    def formula(individu, period):
        return individu("allocation_retour_emploi", period)


class indemnites_chomage_partiel(Variable):
    value_type = float
    entity = Individu
    label = "Indemnités de chômage partiel"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class allocation_retour_emploi(Variable):
    value_type = float
    entity = Individu
    label = "Allocation chômage d'aide au retour à l'emploi (ARE)"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "https://www.legifrance.gouv.fr/codes/id/LEGISCTA000006178163/"


class allocation_retour_emploi_journaliere(Variable):
    value_type = float
    entity = Individu
    label = "Allocation chômage d'aide au retour à l'emploi (ARE) journalière au sens Pôle Emploi"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "https://www.legifrance.gouv.fr/codes/id/LEGISCTA000006178163/"


class allocation_travailleur_independant(Variable):
    value_type = float
    entity = Individu
    label = "Allocation des travailleurs indépendants (ATI)"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    reference = "https://www.legifrance.gouv.fr/codes/id/LEGISCTA000037388330/"
    documentation = '''
    Indemnisation de Pôle emploi en vigueur à partir du 1er novembre 2019 à destination
    des travailleurs non salariés indépendants contraints de mettre fin à leur activité.
    '''



from functools import partial
from numpy import busday_count as original_busday_count, datetime64, timedelta64
from openfisca_france.model.base import *


class indemnites_stage(Variable):
    value_type = float
    entity = Individu
    label = "Indemnités de stage"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class revenus_stage_formation_pro(Variable):
    value_type = float
    entity = Individu
    label = "Revenus de stage de formation professionnelle"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class bourse_recherche(Variable):
    value_type = float
    entity = Individu
    label = "Bourse de recherche"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class sal_pen_exo_etr(Variable):
    cerfa_field = {
        0: "1AC",
        1: "1BC",
        2: "1CC",
        3: "1DC",
        }
    value_type = int
    unit = 'currency'
    entity = Individu
    label = "Salaires et pensions exonérés de source étrangère retenus pour le calcul du taux effectif"
    # start_date = date(2013, 1, 1)
    definition_period = YEAR


class frais_reels(Variable):
    cerfa_field = {
        0: "1AK",
        1: "1BK",
        2: "1CK",
        3: "1DK",
        4: "1EK",
        }
    value_type = int
    unit = 'currency'
    entity = Individu
    label = "Frais réels"
    definition_period = YEAR


class hsup(Variable):
    cerfa_field = {
        0: "1AU",
        1: "1BU",
        2: "1CU",
        3: "1DU",
        }
    value_type = int
    unit = 'currency'
    entity = Individu
    label = "Heures supplémentaires : revenus exonérés connus"
    # start_date = date(2007, 1, 1)
    end = '2013-12-13'
    definition_period = MONTH
    set_input = set_input_divide_by_period
    calculate_output = calculate_output_add


class ppe_du_sa(Variable):
    cerfa_field = {
        0: "1AV",
        1: "1BV",
        2: "1CV",
        3: "1DV",
        4: "1QV",
        }
    value_type = int
    entity = Individu
    label = "Prime pour l'emploi des salariés: nombre d'heures payées"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period):
        heures_remunerees_volume = individu('heures_remunerees_volume', period)
        contrat_travail = individu('contrat_de_travail', period)
        travail_temps_decompte_en_heures = (
            (contrat_travail == TypesContratDeTravail.temps_partiel)
            + (contrat_travail == TypesContratDeTravail.forfait_heures_semaines)
            + (contrat_travail == TypesContratDeTravail.forfait_heures_mois)
            + (contrat_travail == TypesContratDeTravail.forfait_heures_annee)
            + (contrat_travail == TypesContratDeTravail.forfait_jours_annee)
            )

        return heures_remunerees_volume * travail_temps_decompte_en_heures


class ppe_tp_sa(Variable):
    cerfa_field = {
        0: "1AX",
        1: "1BX",
        2: "1CX",
        3: "1DX",
        4: "1QX",
        }
    value_type = bool
    entity = Individu
    label = "Prime pour l'emploi des salariés: indicateur de travail à temps plein sur l'année entière"
    definition_period = YEAR

    def formula(individu, period):
        mois = period.first_month
        indicateur = individu('contrat_de_travail', mois) == 0
        # On parcours tous les mois de l'année pour s'assurer que l'individu était employé à temps plein
        # durant toute l'année.
        while mois.start.month < 12:
            mois = mois.offset(1)
            indicateur = indicateur & (individu('contrat_de_travail', mois) == 0)
        return indicateur


class TypesExpositionAccident(Enum):
    __order__ = 'faible moyen eleve tres_eleve'  # Needed to preserve the enum order in Python 2
    faible = "Faible"
    moyen = "Moyen"
    eleve = "Élevé"
    tres_eleve = "Très élevé"


class exposition_accident(Variable):
    value_type = Enum
    possible_values = TypesExpositionAccident
    default_value = TypesExpositionAccident.faible
    entity = Individu
    label = "Exposition au risque pour les accidents du travail"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class TypesExpositionPenibilite(Enum):
    __order__ = 'nulle simple multiple'  # Needed to preserve the enum order in Python 2
    nulle = "Nulle, pas d'exposition de l'employé à un facteur de pénibilité"
    simple = "Simple, exposition à un seul facteur de pénibilité"
    multiple = "Multiple, exposition à plusieurs facteurs de pénibilité"


class exposition_penibilite(Variable):
    value_type = Enum
    possible_values = TypesExpositionPenibilite
    default_value = TypesExpositionPenibilite.nulle
    entity = Individu
    label = "Exposition à un ou plusieurs facteurs de pénibilité"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class TypesAllegementModeRecouvrement(Enum):
    __order__ = 'fin_d_annee anticipe progressif'  # Needed to preserve the enum order in Python 2
    fin_d_annee = "fin_d_annee"
    anticipe = "anticipe_regularisation_fin_de_periode"
    progressif = "progressif"


class allegement_fillon_mode_recouvrement(Variable):
    value_type = Enum
    possible_values = TypesAllegementModeRecouvrement
    default_value = TypesAllegementModeRecouvrement.fin_d_annee
    entity = Individu
    label = "Mode de recouvrement des allègements Fillon"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class allegement_cotisation_allocations_familiales_mode_recouvrement(Variable):
    value_type = Enum
    possible_values = TypesAllegementModeRecouvrement
    default_value = TypesAllegementModeRecouvrement.fin_d_annee
    entity = Individu
    label = "Mode de recouvrement de l'allègement de la cotisation d'allocations familiales"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class allegement_cotisation_maladie_mode_recouvrement(Variable):
    value_type = Enum
    possible_values = TypesAllegementModeRecouvrement
    default_value = TypesAllegementModeRecouvrement.fin_d_annee
    entity = Individu
    label = "Mode de recouvrement de l'allègement des cotisations maladie sur les bas et moyens salaires (Ex-CICE)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class apprentissage_contrat_debut(Variable):
    value_type = date
    entity = Individu
    label = "Date de début du contrat d'apprentissage"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class arrco_tranche_a_taux_employeur(Variable):
    value_type = float
    entity = Individu
    label = "Taux ARRCO tranche A employeur) propre à l'entreprise"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class arrco_tranche_a_taux_salarie(Variable):
    value_type = float
    entity = Individu
    label = "Taux ARRCO tranche A salarié) propre à l'entreprise"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class assujettie_taxe_salaires(Variable):
    value_type = bool
    entity = Individu
    label = "Entreprise assujettie à la taxe sur les salaires"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class avantage_en_nature_valeur_reelle(Variable):
    value_type = float
    entity = Individu
    label = "Avantages en nature (Valeur réelle)"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class indemnites_compensatrices_conges_payes(Variable):
    value_type = float
    entity = Individu
    label = "indemnites_compensatrices_conges_payes"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class indemnite_fin_contrat_due(Variable):
    value_type = bool
    entity = Individu
    label = "indemnite_fin_contrat_due"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class TypesContratDeTravail(Enum):
    __order__ = 'temps_plein temps_partiel forfait_heures_semaines forfait_heures_mois forfait_heures_annee forfait_jours_annee sans_objet'  # Needed to preserve the enum order in Python 2
    temps_plein = "temps_plein"
    temps_partiel = "temps_partiel"
    forfait_heures_semaines = "forfait_heures_semaines"
    forfait_heures_mois = "forfait_heures_mois"
    forfait_heures_annee = "forfait_heures_annee"
    forfait_jours_annee = "forfait_jours_annee"
    sans_objet = "sans_objet"


class contrat_de_travail(Variable):
    value_type = Enum
    possible_values = TypesContratDeTravail
    default_value = TypesContratDeTravail.temps_plein
    entity = Individu
    label = "Type contrat de travail"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class date_debut_recherche_emploi(Variable):
    value_type = date
    default_value = date(1, 1, 1)
    entity = Individu
    label = "Date de début d'une activité en recherche d'emploi selon Pôle Emploi (exemple : entretien d'embauche, concours public, examen certifiant)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class contrat_de_travail_debut(Variable):
    value_type = date
    default_value = date(1870, 1, 1)
    entity = Individu
    label = "Date d'arrivée dans l'entreprise"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class contrat_de_travail_fin(Variable):
    value_type = date
    default_value = date(2099, 12, 31)
    entity = Individu
    label = "Date de départ de l'entreprise"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class TypesContrat(Enum):
    __order__ = 'aucun cdi cdd ctt formation'  # Needed to preserve the enum order in Python 2
    aucun = "Aucun contrat"
    cdi = "Contrat de travail à durée indéterminée (CDI)"
    cdd = "Contrat de travail à durée déterminée (CDD)"
    ctt = "Contrat de travail temporaire (CTT)"
    formation = "Formation"


class contrat_de_travail_type(Variable):
    value_type = Enum
    possible_values = TypesContrat
    default_value = TypesContrat.cdi
    entity = Individu
    label = "Type du contrat de travail"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class contrat_aide(Variable):
    value_type = bool
    entity = Individu
    label = "L'individu est en contrat aidé"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    reference = "https://dares.travail-emploi.gouv.fr/definitions-et-concepts/contrats-aides"


class duree_formation(Variable):
    value_type = float
    entity = Individu
    label = "Durée de la formation en heures"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class TypesLieuEmploiFormation(Enum):
    non_renseigne = "Non renseigné"
    metropole_hors_corse = "Métropole hors Corse"
    corse = "Corse"
    guadeloupe = "Guadeloupe"
    martinique = "Martinique"
    guyane = "Guyane"
    la_reunion = "La réunion"
    saint_pierre_et_miquelon = "Saint Pierre et Miquelon"
    mayotte = "Mayotte"
    saint_bartelemy = "Saint Bartelemy"
    saint_martin = "Saint Martin"


class lieu_emploi_ou_formation(Variable):
    value_type = Enum
    possible_values = TypesLieuEmploiFormation
    default_value = TypesLieuEmploiFormation.non_renseigne
    entity = Individu
    label = "Zone de l'emploi ou de la formation"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class contrat_de_travail_duree(Variable):
    value_type = float
    default_value = 0.
    entity = Individu
    label = "Durée du contrat de travail en mois"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class TypesCategoriesDemandeurEmploi(Enum):
    # http://www.bo-pole-emploi.org/bulletinsofficiels/instruction-n2016-33-du-6-octobre-2016-bope-n2016-80.html?type=dossiers/2016/bope-n2016-80-du-17-novembre-201
    # https://allocation-chomage.fr/categorie-chomeur/

    __order__ = 'pas_de_categorie categorie_1 categorie_2 categorie_3 categorie_4 categorie_5 categorie_6 categorie_7 categorie_8' \
                # Needed to preserve the enum order in Python 2

    pas_de_categorie = "Aucune catégorie"
    categorie_1 = "Catégorie 1 - Personnes sans emploi, immédiatement disponibles en recherche de CDI plein temps."
    categorie_2 = "Catégorie 2 - Personnes sans emploi, immédiatement disponibles en recherche de CDI à temps partiel."
    categorie_3 = "Catégorie 3 - Personnes sans emploi, immédiatement disponibles en recherche de CDD."
    categorie_4 = "Catégorie 4 - Personnes sans emploi, non immédiatement disponibles et à la recherche d’un emploi."
    categorie_5 = "Catégorie 5 - Personnes non immédiatement disponibles, parce que titulaires d'un ou de plusieurs emplois, et à la recherche d'un autre emploi."
    categorie_6 = "Catégorie 6 - Personnes non immédiatement disponibles, en recherche d'un autre emploi en CDI à plein temps."
    categorie_7 = "Catégorie 7 - Personnes non immédiatement disponibles, en recherche d'un autre emploi en CDI à temps partiel."
    categorie_8 = "Catégorie 8 - Personnes non immédiatement disponibles, en recherche d'un autre emploi en CDD."


class pole_emploi_categorie_demandeur_emploi(Variable):
    reference = [
        "http://www.bo-pole-emploi.org/bulletinsofficiels/instruction-n2016-33-du-6-octobre-2016-bope-n2016-80.html?type=dossiers/2016/bope-n2016-80-du-17-novembre-201",
        "Annexe 3 : la fiche 3 - Les effets de l’inscription"
        ]
    value_type = Enum
    possible_values = TypesCategoriesDemandeurEmploi
    default_value = TypesCategoriesDemandeurEmploi.pas_de_categorie
    entity = Individu
    label = "Le classement des demandeurs d’emploi dans les différentes catégories d’inscription à Pôle Emploi"
    definition_period = MONTH


class TypesCotisationSocialeModeRecouvrement(Enum):
    __order__ = 'mensuel annuel mensuel_strict'  # Needed to preserve the enum order in Python 2
    mensuel = "Mensuel avec régularisation en fin d'année"
    annuel = "Annuel"
    mensuel_strict = "Mensuel strict"


class cotisation_sociale_mode_recouvrement(Variable):
    value_type = Enum
    possible_values = TypesCotisationSocialeModeRecouvrement
    default_value = TypesCotisationSocialeModeRecouvrement.mensuel_strict
    entity = Individu
    label = "Mode de recouvrement des cotisations sociales"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class entreprise_est_association_non_lucrative(Variable):
    value_type = bool
    entity = Individu
    label = "L'entreprise est une association à but non lucratif, par exemple loi de 1901"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class depcom_entreprise(Variable):
    value_type = str
    max_length = 5
    entity = Individu
    label = "Localisation entreprise (depcom)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class code_postal_entreprise(Variable):
    value_type = str
    max_length = 5
    entity = Individu
    label = "Localisation entreprise (Code postal)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class salarie_regime_alsace_moselle(Variable):
    entity = Individu
    value_type = bool
    label = "Le salarié cotise au régime de l'Alsace-Moselle"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    # Attention : ce n'est pas équivalent au fait de travailler en Alsace-Moselle !
    # http://regime-local.fr/salaries/


class effectif_entreprise(Variable):
    entity = Individu
    value_type = int
    label = "Effectif de l'entreprise"
    set_input = set_input_dispatch_by_period
    definition_period = MONTH


class entreprise_assujettie_cet(Variable):
    value_type = bool
    entity = Individu
    label = "Entreprise assujettie à la contribution économique territoriale"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class entreprise_assujettie_is(Variable):
    value_type = bool
    entity = Individu
    label = "Entreprise assujettie à l'impôt sur les sociétés (IS)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class entreprise_benefice(Variable):
    value_type = float
    entity = Individu
    set_input = set_input_divide_by_period
    label = "Bénéfice de l'entreprise"
    definition_period = MONTH
    calculate_output = calculate_output_add


class entreprise_bilan(Variable):
    value_type = float
    entity = Individu
    label = "Bilan de l'entreprise"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class entreprise_chiffre_affaire(Variable):
    value_type = float
    entity = Individu
    label = "Chiffre d'affaire de l'entreprise"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class entreprise_creation(Variable):
    value_type = date
    entity = Individu
    label = "Date de création de l'entreprise"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class nombre_tickets_restaurant(Variable):
    value_type = int
    entity = Individu
    label = "Nombre de tickets restaurant"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class nouvelle_bonification_indiciaire(Variable):
    value_type = float
    entity = Individu
    label = "Nouvelle bonification indicaire"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class prevoyance_obligatoire_cadre_taux_employe(Variable):
    value_type = float
    default_value = 0.015  # 1.5% est le minimum en 2014
    entity = Individu
    label = "Taux de cotisation employeur pour la prévoyance obligatoire des cadres"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class prevoyance_obligatoire_cadre_taux_employeur(Variable):
    value_type = float
    default_value = 0.015  # 1.5% est le minimum en 2014
    entity = Individu
    label = "Taux de cotisation employeur pour la prévoyance obligatoire des cadres"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class primes_salaires(Variable):
    value_type = float
    entity = Individu
    label = "Indemnités, primes et avantages en argent (brut)"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class complementaire_sante_montant(Variable):
    value_type = float
    entity = Individu
    label = "Montant de la complémentaire santé obligatoire retenue par l'employeur"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class complementaire_sante_taux_employeur(Variable):
    value_type = float
    default_value = 0.5
    # La part minimum légale est de 50 %
    entity = Individu
    label = "Part de la complémentaire santé obligatoire payée par l'employeur"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class prise_en_charge_employeur_prevoyance_complementaire(Variable):
    value_type = float
    entity = Individu
    label = "Part salariale des cotisations de prévoyance complémentaire prise en charge par l'employeur"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class prise_en_charge_employeur_retraite_complementaire(Variable):
    value_type = float
    entity = Individu
    label = "Part salariale des cotisations de retraite complémentaire prise en charge par l'employeur"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class prise_en_charge_employeur_retraite_supplementaire(Variable):
    value_type = float
    entity = Individu
    label = "Part salariale des cotisations de retraite supplémentaire prise en charge par l'employeur"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class ratio_alternants(Variable):
    value_type = float
    entity = Individu
    label = "Ratio d'alternants dans l'effectif moyen"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period
    # TODO modéliser cette variable, dont la définition intervient dans le calcul de la Contribution supplémentaire d'apprentissage, et change au cours du temps.


class remboursement_transport_base(Variable):
    value_type = float
    entity = Individu
    label = "Base pour le calcul du remboursement des frais de transport"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class indemnites_forfaitaires(Variable):
    value_type = float
    entity = Individu
    label = "Indemnités forfaitaires (transport, nourriture)"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class salaire_de_base(Variable):
    value_type = float
    entity = Individu
    label = "Salaire de base, en général appelé salaire brut, la 1ère ligne sur la fiche de paie"
    set_input = set_input_divide_by_period
    reference = 'http://www.insee.fr/fr/methodes/default.asp?page=definitions/salaire-mensuel-base-smb.htm'
    definition_period = MONTH


class titre_restaurant_taux_employeur(Variable):
    value_type = float
    default_value = 0.5
    entity = Individu
    label = "Taux de participation de l'employeur au titre restaurant"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class titre_restaurant_valeur_unitaire(Variable):
    value_type = float
    entity = Individu
    label = "Valeur faciale unitaire du titre restaurant"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class titre_restaurant_volume(Variable):
    value_type = int
    entity = Individu
    label = "Volume des titres restaurant"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class traitement_indiciaire_brut(Variable):
    value_type = float
    entity = Individu
    label = "Traitement indiciaire brut (TIB)"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class categorie_salarie(Variable):
    value_type = Enum
    possible_values = TypesCategorieSalarie  # defined in model/base.py
    default_value = TypesCategorieSalarie.prive_non_cadre
    entity = Individu
    label = "Catégorie de salarié"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class heures_duree_collective_entreprise(Variable):
    value_type = int  # TODO default la valeur de la durée légale ?
    entity = Individu
    label = "Durée mensuelle collective dans l'entreprise (heures, temps plein)"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period


class heures_non_remunerees_volume(Variable):
    value_type = float
    entity = Individu
    label = "Volume des heures non rémunérées (convenance personnelle hors contrat/forfait)"
    set_input = set_input_divide_by_period
    definition_period = MONTH


class heures_remunerees_volume(Variable):
    value_type = float
    entity = Individu
    label = "Volume des heures rémunérées contractuellement (heures/mois, temps partiel)"
    set_input = set_input_divide_by_period
    definition_period = MONTH


class forfait_heures_remunerees_volume(Variable):
    value_type = int
    entity = Individu
    label = "Volume des heures rémunérées à un forfait heures"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class forfait_jours_remuneres_volume(Variable):
    value_type = int
    entity = Individu
    label = "Volume des heures rémunérées à forfait jours"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class volume_jours_ijss(Variable):
    value_type = int
    entity = Individu
    label = "Volume des jours pour lesquels sont versés une idemnité journalière par la sécurité sociale"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class avantage_en_nature(Variable):
    value_type = float
    entity = Individu
    label = "Avantages en nature"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        period = period
        avantage_en_nature_valeur_reelle = individu('avantage_en_nature_valeur_reelle', period)
        avantage_en_nature_valeur_forfaitaire = individu('avantage_en_nature_valeur_forfaitaire', period)

        return avantage_en_nature_valeur_reelle + avantage_en_nature_valeur_forfaitaire


class avantage_en_nature_valeur_forfaitaire(Variable):
    value_type = float
    entity = Individu
    label = "Evaluation fofaitaire des avantages en nature "
    definition_period = MONTH
    set_input = set_input_divide_by_period

    # TODO: coplete this function
    def formula(individu, period, parameters):
        period = period
        avantage_en_nature_valeur_reelle = individu('avantage_en_nature_valeur_reelle', period)

        return avantage_en_nature_valeur_reelle * 0


class depense_cantine_titre_restaurant_employe(Variable):
    value_type = float
    entity = Individu
    label = "Dépense de cantine et de titre restaurant à charge de l'employe"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        period = period

        valeur_unitaire = individu("titre_restaurant_valeur_unitaire", period)
        volume = individu("titre_restaurant_volume", period)
        taux_employeur = individu('titre_restaurant_taux_employeur', period)

        return - valeur_unitaire * volume * (1 - taux_employeur)


class depense_cantine_titre_restaurant_employeur(Variable):
    value_type = float
    entity = Individu
    label = "Dépense de cantine et de titre restaurant à charge de l'employeur"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        period = period
        valeur_unitaire = individu("titre_restaurant_valeur_unitaire", period)
        volume = individu("titre_restaurant_volume", period)  # Compute with jours ouvrables ?
        taux_employeur = individu('titre_restaurant_taux_employeur', period)

        return valeur_unitaire * volume * taux_employeur


class nombre_jours_calendaires(Variable):
    value_type = float
    entity = Individu
    label = "Nombre de jours calendaires travaillés"
    definition_period = MONTH
    default_value = 30
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        contrat_de_travail_debut = individu('contrat_de_travail_debut', period)
        contrat_de_travail_fin = individu('contrat_de_travail_fin', period)

        busday_count = partial(original_busday_count, weekmask = "1" * 7)
        debut_mois = datetime64(period.start.offset('first-of', 'month'))
        fin_mois = datetime64(period.start.offset('last-of', 'month'))
        jours_travailles = max_(
            busday_count(
                max_(contrat_de_travail_debut, debut_mois),
                min_(contrat_de_travail_fin, fin_mois) + timedelta64(1, 'D')
                ),
            0,
            )

        return jours_travailles


class remboursement_transport(Variable):
    value_type = float
    entity = Individu
    label = "Remboursement partiel des frais de transport par l'employeur"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):

        remboursement_transport_base = individu('remboursement_transport_base', period)
        # TODO: paramètres en dur dans le code
        return - .5 * remboursement_transport_base


# Fonction publique

class gipa(Variable):
    value_type = float
    entity = Individu
    label = "Indemnité de garantie individuelle du pouvoir d'achat"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    # TODO: à coder


class indemnite_residence(Variable):
    value_type = float
    entity = Individu
    label = "Indemnité de résidence des fonctionnaires"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        traitement_indiciaire_brut = individu('traitement_indiciaire_brut', period)
        salaire_de_base = individu('salaire_de_base', period)
        categorie_salarie = individu('categorie_salarie', period)
        zone_apl = individu.menage('zone_apl', period)
        TypesZoneApl = zone_apl.possible_values
        _P = parameters(period)

        P = _P.prestations_sociales.fonc.indem_resid
        min_zone_1, min_zone_2, min_zone_3 = P.min * P.taux.zone1, P.min * P.taux.zone2, P.min * P.taux.zone3
        taux = P.taux.zone1 * (zone_apl == TypesZoneApl.zone_1) + P.taux.zone2 * (zone_apl == TypesZoneApl.zone_2) + P.taux.zone3 * (zone_apl == TypesZoneApl.zone_3)
        plancher = min_zone_1 * (zone_apl == TypesZoneApl.zone_1) + min_zone_2 * (zone_apl == TypesZoneApl.zone_2) + min_zone_3 * (zone_apl == TypesZoneApl.zone_3)
        public = (
            (categorie_salarie == TypesCategorieSalarie.public_titulaire_etat)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_militaire)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_territoriale)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_hospitaliere)
            + (categorie_salarie == TypesCategorieSalarie.public_non_titulaire)
            )
        return max_(
            plancher,
            taux * (traitement_indiciaire_brut + salaire_de_base)
            ) * public


class indice_majore(Variable):
    value_type = float
    entity = Individu
    label = "Indice majoré"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(individu, period, parameters):
        period = period.start.period('month').offset('first-of')
        categorie_salarie = individu('categorie_salarie', period)
        traitement_indiciaire_brut = individu('traitement_indiciaire_brut', period)
        _P = parameters(period)

        traitement_annuel_brut = _P.prestations_sociales.fonc.IM_100
        public = (
            (categorie_salarie == TypesCategorieSalarie.public_titulaire_etat)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_militaire)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_territoriale)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_hospitaliere)
            )

        return (traitement_indiciaire_brut * 100 * 12 / traitement_annuel_brut) * public


class primes_fonction_publique(Variable):
    value_type = float
    entity = Individu
    label = "Calcul des primes pour les fonctionnaries"
    reference = "http://vosdroits.service-public.fr/particuliers/F465.xhtml"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        # period = period.first_month
        categorie_salarie = individu('categorie_salarie', period)
        traitement_indiciaire_brut = individu('traitement_indiciaire_brut', period)

        public = (
            (categorie_salarie == TypesCategorieSalarie.public_titulaire_etat)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_militaire)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_territoriale)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_hospitaliere)
            )

        return TAUX_DE_PRIME * traitement_indiciaire_brut * public


class af_nbenf_fonc(Variable):
    value_type = int
    label = "Nombre d'enfants dans la famille au sens des allocations familiales pour les fonctionnaires"
    entity = Famille
    definition_period = MONTH
    set_input = set_input_dispatch_by_period

    def formula(famille, period, parameters):
        """
            Cette variable est une version légèrement modifiée de `af_nbenf`. Elle se base sur le salaire de base, tandis que `af_nbenf` se base sur le salaire net.
            On ne peut pas utiliser la variable `af_nbenf` dans le calcul de `supplement_familial_traitement` (ci-dessous) car `af_nbenf` dépend du `salaire_net`, et `salaire_net` dépends de `supplement_familial_traitement`. Cela créerait une boucle infinie.
            D'où l'introduction de cette variable alternative.
        """

        salaire_de_base_mensualise = famille.members('salaire_de_base', period.start.period('month', 6).offset(-6), options = [ADD])
        law = parameters(period)
        nbh_travaillees = 169
        smic_mensuel_brut = law.marche_travail.salaire_minimum.smic.smic_b_horaire * nbh_travaillees

        autonomie_financiere = (salaire_de_base_mensualise >= (
            law.prestations_sociales.prestations_familiales.def_pac.revenu_plafond_pac_non_scolaire
            * smic_mensuel_brut
            ))

        age = famille.members('age', period)

        condition_enfant = (
            (age >= law.prestations_sociales.prestations_familiales.prestations_generales.af.af_cm.age1)
            * (age <= law.prestations_sociales.prestations_familiales.prestations_generales.af.af_cm.age2)
            * not_(autonomie_financiere)
            )

        return famille.sum(condition_enfant, role = Famille.ENFANT)


class supplement_familial_traitement(Variable):
    value_type = float
    entity = Individu
    label = "Supplément familial de traitement"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    # Attention : par hypothèse ne peut êre attribué qu'à la tête du ménage
    # TODO: gérer le cas encore problématique du conjoint fonctionnaire

    def formula(individu, period, parameters):
        categorie_salarie = individu('categorie_salarie', period)
        traitement_indiciaire_brut = individu('traitement_indiciaire_brut', period)
        _P = parameters(period)

        fonc_nbenf = individu.famille('af_nbenf_fonc', period) * individu.has_role(Famille.DEMANDEUR)

        P = _P.prestations_sociales.fonc.supplement_familial
        part_fixe_1 = P.fixe.enf1
        part_fixe_2 = P.fixe.enf2
        part_fixe_supp = P.fixe.enfsupp

        part_fixe = (
            part_fixe_1 * (fonc_nbenf == 1)
            + part_fixe_2 * (fonc_nbenf == 2)
            + part_fixe_supp * max_(0, fonc_nbenf - 2)
            )

        # pct_variable_1 = 0
        pct_variable_2 = P.prop.enf2
        pct_variable_3 = P.prop.enf3
        pct_variable_supp = P.prop.enfsupp

        pct_variable = (
            pct_variable_2 * (fonc_nbenf == 2)
            + (pct_variable_3) * (fonc_nbenf == 3)
            + pct_variable_supp * max_(0, fonc_nbenf - 3)
            )

        indice_maj_min = P.IM_min
        indice_maj_max = P.IM_max

        traitement_brut_mensuel_min = _traitement_brut_mensuel(indice_maj_min, _P.prestations_sociales.fonc.IM_100)
        plancher_mensuel_1 = part_fixe
        plancher_mensuel_2 = part_fixe + traitement_brut_mensuel_min * pct_variable_2
        plancher_mensuel_3 = part_fixe + traitement_brut_mensuel_min * pct_variable_3
        plancher_mensuel_supp = traitement_brut_mensuel_min * pct_variable_supp

        plancher = (
            plancher_mensuel_1 * (fonc_nbenf == 1)
            + plancher_mensuel_2 * (fonc_nbenf == 2)
            + plancher_mensuel_3 * (fonc_nbenf >= 3)
            + plancher_mensuel_supp * max_(0, fonc_nbenf - 3)
            )

        traitement_brut_mensuel_max = _traitement_brut_mensuel(indice_maj_max, _P.prestations_sociales.fonc.IM_100)
        plafond_mensuel_1 = part_fixe
        plafond_mensuel_2 = part_fixe + traitement_brut_mensuel_max * pct_variable_2
        plafond_mensuel_3 = part_fixe + traitement_brut_mensuel_max * pct_variable_3
        plafond_mensuel_supp = traitement_brut_mensuel_max * pct_variable_supp

        plafond = (
            plafond_mensuel_1 * (fonc_nbenf == 1)
            + plafond_mensuel_2 * (fonc_nbenf == 2)
            + plafond_mensuel_3 * (fonc_nbenf == 3)
            + plafond_mensuel_supp * max_(0, fonc_nbenf - 3)
            )

        public = (
            (categorie_salarie == TypesCategorieSalarie.public_titulaire_etat)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_militaire)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_territoriale)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_hospitaliere)
            + (categorie_salarie == TypesCategorieSalarie.public_non_titulaire)
            )

        sft = public * min_(
            max_(part_fixe + pct_variable * traitement_indiciaire_brut, plancher),
            plafond
            )

        return sft


def _traitement_brut_mensuel(indice_maj, indice_majore_100_annuel):
    traitement_brut = indice_majore_100_annuel * indice_maj / 100 / 12
    return traitement_brut


class remuneration_principale(Variable):
    value_type = float
    entity = Individu
    label = "Rémunération principale des agents titulaires de la fonction publique"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        traitement_indiciaire_brut = individu('traitement_indiciaire_brut', period)
        nouvelle_bonification_indiciaire = individu('nouvelle_bonification_indiciaire', period)
        categorie_salarie = individu('categorie_salarie', period)

        public = (
            (categorie_salarie == TypesCategorieSalarie.public_titulaire_etat)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_militaire)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_territoriale)
            + (categorie_salarie == TypesCategorieSalarie.public_titulaire_hospitaliere)
            )

        return (
            public * (
                traitement_indiciaire_brut + nouvelle_bonification_indiciaire
                )
            )


class salaire_net_a_payer(Variable):
    value_type = float
    entity = Individu
    label = "Salaire net à payer (fiche de paie)"
    set_input = set_input_divide_by_period
    definition_period = MONTH

    def formula(individu, period, parameters):
        '''
        Calcul du salaire net à payer après déduction des sommes
        dues par les salarié avancées par l'employeur
        '''
        salaire_net = individu('salaire_net', period, options = [ADD])
        depense_cantine_titre_restaurant_employe = individu(
            'depense_cantine_titre_restaurant_employe', period)
        indemnites_forfaitaires = individu('indemnites_forfaitaires', period)
        remuneration_apprenti = individu('remuneration_apprenti', period)
        stage_gratification = individu('stage_gratification', period)
        salaire_net_a_payer = (
            salaire_net
            + remuneration_apprenti
            + stage_gratification
            + depense_cantine_titre_restaurant_employe
            + indemnites_forfaitaires
            )
        return salaire_net_a_payer


class salaire_super_brut_hors_allegements(Variable):
    value_type = float
    entity = Individu
    label = "Salaire super-brut (fiche de paie): rémunération + cotisations sociales employeur"
    set_input = set_input_divide_by_period
    definition_period = MONTH

    def formula(individu, period, parameters):
        salaire_de_base = individu('salaire_de_base', period)
        remuneration_principale = individu('remuneration_principale', period)
        remuneration_apprenti = individu('remuneration_apprenti', period)

        primes_fonction_publique = individu('primes_fonction_publique', period)
        indemnite_residence = individu('indemnite_residence', period)
        supplement_familial_traitement = individu('supplement_familial_traitement', period)
        cotisations_employeur = individu('cotisations_employeur', period)
        depense_cantine_titre_restaurant_employeur = individu('depense_cantine_titre_restaurant_employeur', period)
        reintegration_titre_restaurant_employeur = individu('reintegration_titre_restaurant_employeur', period)
        indemnite_fin_contrat = individu('indemnite_fin_contrat', period)
        salaire_super_brut_hors_allegements = (
            salaire_de_base
            + remuneration_apprenti
            + indemnite_fin_contrat
            + remuneration_principale
            + primes_fonction_publique
            + indemnite_residence
            + supplement_familial_traitement
            + depense_cantine_titre_restaurant_employeur
            - reintegration_titre_restaurant_employeur
            - cotisations_employeur
            )

        return salaire_super_brut_hors_allegements


class salaire_super_brut(Variable):
    value_type = float
    entity = Individu
    label = "Coût du travail à court terme. Inclut les exonérations et allègements de charges"
    set_input = set_input_divide_by_period
    definition_period = MONTH

    def formula(individu, period, parameters):
        period = period
        salaire_super_brut_hors_allegements = individu('salaire_super_brut_hors_allegements', period)
        exonerations_et_allegements = individu('exonerations_et_allegements', period)

        return salaire_super_brut_hors_allegements - exonerations_et_allegements


class exonerations_et_allegements(Variable):
    value_type = float
    entity = Individu
    label = "Exonérations et allègements"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        exoneration_cotisations_employeur_apprenti = individu(
            'exoneration_cotisations_employeur_apprenti', period, options = [ADD])
        exoneration_cotisations_employeur_geographiques = individu(
            'exoneration_cotisations_employeur_geographiques', period)
        exoneration_cotisations_employeur_jei = individu(
            'exoneration_cotisations_employeur_jei', period, options = [ADD])
        exoneration_cotisations_employeur_stagiaire = individu(
            'exoneration_cotisations_employeur_stagiaire', period, options = [ADD])

        allegement_fillon = individu('allegement_fillon', period, options = [ADD])
        allegement_cotisation_maladie = individu('allegement_cotisation_maladie', period, options = [ADD])
        allegement_cotisation_allocations_familiales = individu('allegement_cotisation_allocations_familiales', period, options = [ADD])

        return (
            allegement_fillon
            + allegement_cotisation_maladie
            + allegement_cotisation_allocations_familiales
            + exoneration_cotisations_employeur_geographiques
            + exoneration_cotisations_employeur_jei
            + exoneration_cotisations_employeur_apprenti
            + exoneration_cotisations_employeur_stagiaire
            )


class cout_du_travail(Variable):
    value_type = float
    entity = Individu
    label = "Coût du travail à long terme. Inclut les charges, aides et crédits différés"
    set_input = set_input_divide_by_period
    definition_period = MONTH
    calculate_output = calculate_output_add

    def formula(individu, period, parameters):
        salaire_super_brut = individu('salaire_super_brut', period)
        cout_differe = individu('cout_differe', period)

        return salaire_super_brut - cout_differe


class cout_differe(Variable):
    value_type = float
    entity = Individu
    label = "Charges, aides et crédits différés ou particuliers"
    definition_period = MONTH
    set_input = set_input_divide_by_period

    def formula(individu, period, parameters):
        credit_impot_competitivite_emploi = individu('credit_impot_competitivite_emploi', period, options = [ADD])
        aide_premier_salarie = individu('aide_premier_salarie', period, options = [ADD])
        aide_embauche_pme = individu('aide_embauche_pme', period, options = [ADD])
        tehr = individu('tehr', period, options = [DIVIDE])

        return credit_impot_competitivite_emploi + aide_premier_salarie + aide_embauche_pme + tehr


class TypesConges(Enum):
    non_renseigne = "Non renseigné"
    conge_parental = "Congé parental"
    conge_maternite_paternite = "Congé maternité ou paternité"
    conge_presence_parentale = "Congé de présence parentale"
    conge_conventionnel = "Congé conventionnel"
    conge_sans_solde = "Congé sans solde"
    disponibilite = "Mise en disponibilité (fonction publique)"
    conge_sabbatique = "Congé sabbatique"


class type_conges(Variable):
    value_type = Enum
    possible_values = TypesConges
    default_value = TypesConges.non_renseigne
    entity = Individu
    label = "Type de congés en cours"
    definition_period = MONTH
    set_input = set_input_dispatch_by_period



#! /usr/bin/env python
import time
from openfisca_core.simulation_builder import SimulationBuilder
from openfisca_france import FranceTaxBenefitSystem
from openfisca_france.situation_examples import couple

tax_benefit_system = FranceTaxBenefitSystem()

NB_RUN = 10

situation = {
    "individus": {
        "demandeur": {
            "salaire_imposable": {
                "2017-12": 62191,
                "2018-01": 62191,
                "2018-02": 62191,
                "2018-03": 62191,
                "2018-04": 62191,
                "2018-05": 62191,
                "2018-06": 62191,
                "2018-07": 62191,
                "2018-08": 62191,
                "2018-09": 62191,
                "2018-10": 62191,
                "2018-11": 62191,
                "2018-12": 62191,
                "2016-01": 62191,
                "2016-02": 62191,
                "2016-03": 62191,
                "2016-04": 62191,
                "2016-05": 62191,
                "2016-06": 62191,
                "2016-07": 62191,
                "2016-08": 62191,
                "2016-09": 62191,
                "2016-10": 62191,
                "2016-11": 62191,
                "2016-12": 62191
                }
            }
        },
    "familles": {
        "_": {
            "parents": [
                "demandeur"
                ]
            }
        },
    "foyers_fiscaux": {
        "_": {
            "declarants": [
                "demandeur"
                ],
            "personnes_a_charge": []
            }
        },
    "menages": {
        "_": {
            "personne_de_reference": [
                "demandeur"
                ]
            }
        }
    }


def timeit(method):
    def timed(*args, **kwargs):
        start_time = time.time()
        result = method(*args, **kwargs)
        delta = time.time() - start_time
        return result, delta

    return timed


def test_revenu_disponible():
    def new_simulation():
        return SimulationBuilder().build_from_dict(tax_benefit_system, couple)

    simulations = [new_simulation() for i in range(NB_RUN)]

    @timeit
    def run_test():
        for simulation in simulations:
            simulation.calculate('revenu_disponible', 2018)

    result, delta = run_test()
    print('{:2.6f} s'.format(delta / NB_RUN))  # noqa T001


def test_spiral():
    def new_simulation():
        return SimulationBuilder().build_from_dict(tax_benefit_system, situation)

    simulations = [new_simulation() for i in range(NB_RUN)]

    @timeit
    def run_test():
        for simulation in simulations:
            simulation.calculate('ass', '2018-12')
            simulation.calculate('logement_social_eligible', '2018-12')

    result, delta = run_test()
    print('{:2.6f} s'.format(delta / NB_RUN))  # noqa T001


print('Premier test revenu disponible')  # noqa T001
test_revenu_disponible()
print('Second test revenu disponible')  # noqa T001
test_revenu_disponible()
print('3e test revenu disponible')  # noqa T001
test_revenu_disponible()

print('Premier test ciblé spirale')  # noqa T001
test_spiral()
print('Second test ciblé spirale')  # noqa T001
test_spiral()
print('3e test ciblé spirale')  # noqa T001
test_spiral()



#! /usr/bin/env python

from setuptools import setup, find_packages
from pathlib import Path

# Read the contents of our README file for PyPi
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name = "OpenFisca-France",
    version = "110.0.2",
    author = "OpenFisca Team",
    author_email = "contact@openfisca.fr",
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Information Analysis",
        ],
    description = "French tax and benefit system for OpenFisca",
    keywords = "benefit france microsimulation social tax",
    license = "http://www.fsf.org/licensing/licenses/agpl-3.0.html",
    license_files = ("LICENSE.AGPL.txt",),
    url = "https://github.com/openfisca/openfisca-france",
    long_description=long_description,
    long_description_content_type='text/markdown',

    data_files = [
        (
            "share/openfisca/openfisca-france",
            ["CHANGELOG.md", "README.md"],
            ),
        ],
    extras_require = {
        "inversion_revenus": [
            "scipy >= 0.17",
            ],
        "de_net_a_brut": [
            "scipy >= 0.17",
            ],
        "taxipp": [
            "pandas >= 0.13",
            ],
        "dev": [
            "autopep8 ==1.5.7",
            "flake8 >=3.8.0,<3.10.0",
            "flake8-print",
            "pytest >= 5.0.0, < 7.0.0",
            "scipy >= 0.17",  # Only used to test de_net_a_brut reform
            "requests >= 2.8",
            "yamllint >=1.11.1,<1.27"
            ],
        "casd-dev": [
            # Same as dev with packages not available at CASD removed
            "autopep8 >=1.3.2",
            "pytest >= 5.0.0, < 7.0.0",
            "requests >= 2.8",
            "scipy >= 0.17",  # Only used to test de_net_a_brut reform
            ]
        },
    include_package_data = True,  # Will read MANIFEST.in
    install_requires = [
        "OpenFisca-Core >=35.2.0,<36.0",
        ],
    packages = find_packages(exclude = [
        "openfisca_france.tests*",
        "openfisca_france.assets.taxe_habitation.source*",
        ]),
    )



from openfisca_france.model.base import *


class pensions_alimentaires_percues(Variable):
    cerfa_field = {
        0: "1AO",
        1: "1BO",
        2: "1CO",
        3: "1DO",
        4: "1EO",
        }
    value_type = float
    unit = 'currency'
    entity = Individu
    label = "Pensions alimentaires perçues"
    definition_period = MONTH
    set_input = set_input_divide_by_period
    calculate_output = calculate_output_add


class pensions_alimentaires_percues_decl(Variable):
    value_type = bool
    default_value = True
    entity = Individu
    label = "Pension déclarée"
    definition_period = YEAR


class pensions_alimentaires_versees_individu(Variable):
    value_type = float
    entity = Individu
    label = "Pensions alimentaires versées pour un individu"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class gains_exceptionnels(Variable):
    value_type = float
    entity = Individu
    label = "Gains exceptionnels"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class allocation_securisation_professionnelle(Variable):
    value_type = float
    entity = Individu
    label = "Allocation de sécurisation professionnelle"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class prime_forfaitaire_mensuelle_reprise_activite(Variable):
    value_type = float
    entity = Individu
    label = "Prime forfaitaire mensuelle pour la reprise d'activité"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class indemnites_volontariat(Variable):
    value_type = float
    entity = Individu
    label = "Indemnités de volontariat"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class dedommagement_victime_amiante(Variable):
    value_type = float
    entity = Individu
    label = "Dédommagement versé aux victimes de l'amiante"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class prestation_compensatoire(Variable):
    value_type = float
    entity = Individu
    label = "Prestation compensatoire"
    definition_period = MONTH
    set_input = set_input_divide_by_period


class pensions_invalidite(Variable):
    value_type = float
    entity = Individu
    label = "Pensions d'invalidité"
    # Cette case est apparue dans la déclaration 2014
    # Auparavant, les pensions d'invalidité étaient incluses dans la case 1AS
    cerfa_field = {
        0: "1AZ",
        1: "1BZ",
        2: "1CZ",
        3: "1DZ",
        }
    # start_date = date(2014, 1, 1)
    definition_period = MONTH
    set_input = set_input_divide_by_period


class bourse_enseignement_sup(Variable):
    value_type = float
    entity = Individu
    label = "Toute bourse de l'enseignement supérieur"
    definition_period = MONTH
    set_input = set_input_divide_by_period


# Avoir fiscaux et crédits d'impôt
# f2ab déjà disponible
class f8ta(Variable):
    cerfa_field = "8TA"
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Retenue à la source en France ou impôt payé à l'étranger"
    definition_period = YEAR


class f8vl(Variable):
    cerfa_field = "8VL"
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Impôt payé à l'étranger sur revenus de capitaux mobiliers et plus-values ouvrant droit à un crédit d'impôt"
    # start_date = date(2016, 1, 1)
    definition_period = YEAR


class f8vm(Variable):
    cerfa_field = {
        0: "8VM",
        1: "8WM",
        2: "8UM",
        }
    value_type = int
    unit = 'currency'
    entity = Individu
    label = "Impôt payé à l'étranger sur revenus de capitaux mobiliers et plus-values ouvrant droit à un crédit d'impôt"
    # start_date = date(2016, 1, 1)
    definition_period = YEAR


class f8th(Variable):
    cerfa_field = "8TH"
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Retenue à la source élus locaux"
    definition_period = YEAR


class f8td_2002_2005(Variable):
    cerfa_field = "8TD"
    value_type = int
    entity = FoyerFiscal
    label = "Contribution exceptionnelle sur les hauts revenus"
    # start_date = date(2002, 1, 1)
    end = '2005-12-31'
    definition_period = YEAR


class f8td(Variable):
    cerfa_field = "8TD"
    value_type = bool
    entity = FoyerFiscal
    label = "Revenus non imposables dépassent la moitié du RFR"
    # start_date = date(2011, 1, 1)
    end = '2014-12-31'
    definition_period = YEAR


class f8ti(Variable):
    cerfa_field = "8TK"
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Revenus de l'étranger exonérés d'impôt"
    definition_period = YEAR


# Auto-entrepreneur : versements libératoires d’impôt sur le revenu
class f8uy(Variable):
    cerfa_field = "8UY"
    value_type = int
    unit = 'currency'
    entity = FoyerFiscal
    label = "Auto-entrepreneur : versements libératoires d’impôt sur le revenu dont le remboursement est demandé"
    # start_date = date(2009, 1, 1)
    definition_period = YEAR
